from elasticsearch import ElasticsearchException
from toflerdb.utils.common import Common
from toflerdb.config import SNAPSHOT_INDEX, SNAPSHOT_DOC_TYPE


def execute_query(query, **kwargs):
    es = Common.get_elasticsearch().get_connection()
    try:
        response = es.search(
            index=SNAPSHOT_INDEX, doc_type=SNAPSHOT_DOC_TYPE,
            body=query, **kwargs)
        return response
    except ElasticsearchException, e:
        Common.get_logger().error('Error ES query execution\n%s' % e)
    except ValueError, e:
        Common.log_error('Invalid response status received from elastic')


def execute_search_query(query, **kwargs):
    es = Common.get_elasticsearch().get_connection()
    try:
        response = es.search(
            index=SNAPSHOT_INDEX, doc_type=SNAPSHOT_DOC_TYPE,
            body=query, **kwargs)
        return response
    except ElasticsearchException, e:
        Common.get_logger().error('Error ES query execution\n%s' % str(e))


def insert_into_snapshot(inputs):
    data = []
    for item in inputs:
        data.append(inputs[item])
    Common.get_elasticsearch().put_bulk_data(
        SNAPSHOT_INDEX, SNAPSHOT_DOC_TYPE, data)


def delete_snapshot_nodes_by_id(id_list):
    if not isinstance(id_list, list):
        id_list = [id_list]

    es = Common.get_elasticsearch().get_connection()
    for i in id_list:
        try:
            es.delete(index=SNAPSHOT_INDEX, doc_type=SNAPSHOT_DOC_TYPE, id=i)
        except ElasticsearchException, e:
            Common.get_logger().error('Error deleting node<%s>\n%s' % (i, e))


def get_snapshot_nodes(**kwargs):
    for key, val in kwargs.iteritems():
        break

    if not isinstance(val, list):
        val = [val]
    es = Common.get_elasticsearch_connection()
    q_string_arr = []
    for elem in val:
        q_string_arr.append({
            "query_string": {
                "fields": [key, "*.%s" % key],
                "query": elem
            }
        })
    try:
        query = {
            "query": {
                "bool": {
                    "should": q_string_arr
                }
            }
        }
        response = es.search(index=SNAPSHOT_INDEX,
                             doc_type=SNAPSHOT_DOC_TYPE, body=query)
        entities = {}
        if response['hits']['total'] > 0:
            for hit in response['hits']['hits']:
                entities[hit['_id']] = hit['_source']
            return entities
        return None
    except Exception as e:
        logger = Common.get_logger()
        logger.error(str(e))
        return None


def find_nodes_with_incoming_references(nodeid):
    query = {"query": {"match": {"_all": nodeid}}}
    es = Common.get_elasticsearch_connection()
    response = es.search(index=SNAPSHOT_INDEX,
                         doc_type=SNAPSHOT_DOC_TYPE, body=query)
    # response = response[1]
    if response['hits']['total'] < 1:
        return []
    hits = response['hits']['hits']
    return hits


def get_related_nodes(nodeid):
    query = {"query": {"match": {"id": nodeid}}}
    es = Common.get_elasticsearch_connection()
    response = es.search(index=SNAPSHOT_INDEX,
                         doc_type=SNAPSHOT_DOC_TYPE, body=query)
    # response = response[1]
    if response['hits']['total'] < 1:
        return None
    node = response['hits']['hits'][0]
    references = _gather_references(node['_source'])
    if len(references) == 0:
        return None
    match_segments = []
    for ref in references:
        match_segments.append({"match": {"id": ref[1]}})

    query = {"query": {"bool": {"should": match_segments}}}
    response = es.search(index=SNAPSHOT_INDEX,
                         doc_type=SNAPSHOT_DOC_TYPE, body=query)
    # response = response[1]
    if response['hits']['total'] < 1:
        return None
    return {'relationships': references, 'nodes': response['hits']['hits']}


def _gather_references(node, field_name=None):
    references = []
    for f in node:
        if field_name is None:
            new_field_name = f
        else:
            new_field_name = "%s.%s" % (field_name, f)
        if f == 'value':
            if type(node[f]) is long or type(node[f]) is int:
                references.append((field_name, node[f]))
        elif type(node[f]) is list:
            for item in node[f]:
                references += _gather_references(item, new_field_name)
        elif type(node[f]) is dict:
            references += _gather_references(node[f], new_field_name)
    return references


def _create_freetextsearch_output(data, request, output):
    for f in request:
        if type(request[f]) is dict:
            if f not in data:
                continue
            op = {}
            if type(data[f]) is list:
                new_array = []
                for item in data[f]:
                    item.pop('fact_id')
                    new_array.append(item)
                _create_freetextsearch_output(new_array, request[f], op)
            else:
                data[f].pop('fact_id')
                _create_freetextsearch_output(data[f], request[f], op)
            output[f] = op
        else:
            if f in data:
                if type(data[f]) is list:
                    new_array = []
                    for item in data[f]:
                        new_array.append(item['value'])
                    output[f] = new_array
                else:
                    if f == 'id':
                        output[f] = data[f]
                    else:
                        output[f] = data[f]['value']
    if 'id' in output:
        output['id'] = str(output['id'])


def freetextsearch(query):
    elastic_query = {"query": {"match": {"_all": query}}}
    es = Common.get_elasticsearch_connection()
    response = es.search(index=SNAPSHOT_INDEX,
                         doc_type=SNAPSHOT_DOC_TYPE, body=elastic_query)
    output = []
    hits = response['hits']['hits']
    for hit in hits:
        src = hit['_source']
        opsrc = {}
        _create_freetextsearch_output(src, src, opsrc)
        output.append(opsrc)
    return output


def general_search(text):
    elastic_query = {
        "query": {
            "bool": {
                "should": [
                    {"match": {"to:label.value": text}},
                    {"match": {"to:label.value.value": text}},
                    {"match": {"to:label.value.whitespace_edgegram": text}},
                    {"match": {"to:label.value.whitespace_keyword": text}}
                ],
                "minimum_should_match": 1
            }
        }
    }
    response = execute_search_query(elastic_query, size=25)
    retval = []
    for hit in response['hits']['hits']:
        src = hit['_source']
        item = {}
        item['id'] = src['id']
        item['to:label'] = []
        for label in src['to:label']:
            item['to:label'].append(label['value'])
        retval.append(item)

    return retval



import json
import os


DEFAULT_CONFIG = {
    "db": {
        "host": "localhost",
        "db": "toflerdb",
        "user": "user",
        "passwd": "password",
        "pool_size": 1
    },
    "elastic": {
        "host": "localhost",
        "port": 9200,
        "index": "toflerdb_snapshot",
        "doc_type": "node"
    },
    "redis": {
        "host": "localhost",
        "port": 6379
    }
}


def run():
    home = os.getenv('HOME')
    location = os.path.join(home, '.toflerdb.conf')
    if not os.path.isfile(location):
        f = open(location, 'w')
        f.write(json.dumps(DEFAULT_CONFIG, indent=4))
        f.close()
    else:
        print "File %s already exists" % location


if __name__ == '__main__':
    run()



import StringIO
import sys
# from toflerdb.utils.common import Common
from toflerdb.dbutils import eternity as eternity_dbutils
from toflerdb.utils import rdfreader, collection, exceptions, templatizedid
from onto_maker import OntoMaker
from entry_maker import EntryMaker
from eraser import Eraser
from graphquery import GraphQuery


def draw_progress_bar(percent, barlen=40):
    print "\r",
    progress = ""
    for i in range(barlen):
        if i < int(barlen * percent):
            progress += "="
        else:
            progress += " "
    print "[ %s ] %.2f%%" % (progress, percent * 100),
    sys.stdout.flush()


def _gather_fact_info(fact, author=None):
    '''
    this is method finds fact_id given fact_tuple or vice versa
    '''
    if isinstance(fact, tuple):
        fact_tuple = fact
        fact_id = eternity_dbutils.get_fact_ids([fact], author)
        if len(fact_id):
            fact_id = fact_id[0]
    else:
        fact_id = fact
        fact_tuple = eternity_dbutils.get_fact_tuples_by_fact_ids(fact_id)
        if len(fact_tuple):
            fact_tuple = fact_tuple[0]

    return (fact_id, fact_tuple)


def _insert_ontology_from_file_stream(
        file_stream, author=None, validation=True):
    onto_tuples = []
    reader = rdfreader.get_rdf_reader(file_stream)
    curr_line = 0
    for row in reader:
        curr_line += 1
        if collection.is_empty_list(row):
            continue
        if row[0].startswith('#'):
            continue
        if row[0].startswith('START_TRANSACTION'):
            continue
        if row[0].startswith('COMMIT'):
            _insert_ontology_from_tuples(
                onto_tuples, author=author, validation=validation)
            onto_tuples = []
            continue
        onto_tuples.append((row[0], row[1], row[2]))


def _insert_ontology_from_file_text(
        file_text, author=None, validation=None):
    file_text = collection.adjust_start_commit_text(file_text)

    csvfile = StringIO.StringIO(file_text)
    _insert_ontology_from_file_stream(
        csvfile, author=author, validation=validation)


def _insert_ontology_from_file(
        file, author=None, validation=True):
    with open(file, 'rb') as csvfile:
        _insert_ontology_from_file_stream(
            csvfile, author=author, validation=validation)


def _insert_ontology_from_tuples(onto_tuples, author=None, validation=True):
    om = OntoMaker()
    # currently validation is not supported in ontology
    # so we need to turn off validation always
    if validation:
        om.turn_on_validation()
    else:
        om.turn_off_validation()
    for row in onto_tuples:
        om.add_input(row[0], row[1], row[2])

    om.make_mapping()
    om.commit()


def insert_ontology(onto_tuples=None, file=None, file_text=None,
                    author=None, validation=True):
    if onto_tuples:
        _insert_ontology_from_tuples(onto_tuples, author, validation)
    if file:
        _insert_ontology_from_file(file, author, validation)
    if file_text:
        _insert_ontology_from_file_text(
            file_text, author, validation)


def _insert_facts_from_tuples(
        fact_tuples, author=None, validation=True, **kwargs):
    if not isinstance(fact_tuples, list):
        fact_tuples = [fact_tuples]

    ignore_duplicate = kwargs.get('ignore_duplicate', True)
    em = EntryMaker(author=author, ignore_duplicate=ignore_duplicate)
    if validation:
        em.turn_on_validation()
    else:
        em.turn_off_validation()

    for fact_tuple in fact_tuples:
        em.add_input(*fact_tuple)

    em.commit()
    response = {}
    response['templatized_id_map'] = templatizedid.trim_userid(
        em._templatized_id_map, author)
    return response


def _insert_facts_from_file_stream(
        file_stream, author=None, validation=True, **kwargs):
    response = {}
    templatized_id_map = {}
    fact_tuples = []
    reader = rdfreader.get_rdf_reader(file_stream)
    line_num = 0
    if kwargs.get('show_progress'):
        reader = list(reader)
        total_len = len(reader)

    for row in reader:
        line_num += 1
        if kwargs.get('start_pos') and line_num < kwargs['start_pos']:
            continue
        if collection.is_empty_list(row):
            continue
        if row[0].startswith('#'):
            continue
        if row[0].startswith('START_TRANSACTION'):
            continue
        if row[0].startswith('COMMIT'):
            if kwargs.get('show_progress'):
                draw_progress_bar((line_num * 1.0) / total_len)
            res = _insert_facts_from_tuples(
                fact_tuples, author=author, validation=validation, **kwargs)
            templatized_id_map.update(res['templatized_id_map'])
            response.update(res)
            if kwargs.get('facts_did_commit'):
                commit_info = {}
                commit_info['commit_pos'] = line_num
                commit_info.update(kwargs.get('callback_info', {}))
                kwargs['facts_did_commit'](commit_info)
            fact_tuples = []
            continue
        fact_tuples.append((row[0], row[1], row[2]))

    response['templatized_id_map'] = templatized_id_map
    return response


def _insert_facts_from_file_text(
        file_text, author=None, validation=None, **kwargs):
    file_text = collection.adjust_start_commit_text(file_text)
    csvfile = StringIO.StringIO(file_text)
    return _insert_facts_from_file_stream(
        csvfile, author=author, validation=validation, **kwargs)


def _insert_facts_from_file(file, author=None, validation=None, **kwargs):
    with open(file, 'rb') as csvfile:
        return _insert_facts_from_file_stream(
            csvfile, author=author, validation=validation, **kwargs)


def _insert_facts_from_s3_file(s3_key, author=None, validation=True, **kwargs):
    # file_text = Common.get_document_from_s3(s3_key)
    csvfile = StringIO.StringIO(file_text)
    response = _insert_facts_from_file_stream(
        csvfile, author=author, validation=validation, **kwargs)
    csvfile.close()
    return response


def insert_facts(
        fact_tuples=None, file=None, file_text=None,
        s3=None, author=None, validation=True, **kwargs):
    if not author:
        raise exceptions.MissingRequiredFieldError(
            'Inserting fact without author key is not allowed')

    if fact_tuples:
        return _insert_facts_from_tuples(
            fact_tuples, author, validation, **kwargs)
    if file:
        return _insert_facts_from_file(file, author, validation, **kwargs)
    if file_text:
        return _insert_facts_from_file_text(
            file_text, author, validation, **kwargs)
    if s3:
        return _insert_facts_from_s3_file(s3, author, validation, **kwargs)


def _update_facts_from_value_map(fact_values, author=None):
    fact_ids = []
    fact_tuples = []
    for fact in fact_values:
        (fact_id, fact_tuple) = _gather_fact_info(fact, author)
        fact_ids.append(fact_id)
        fact_tuples.append(
            (fact_tuple[0], fact_tuple[1], fact_values[fact], fact_id))

    erase_facts(fact_ids, author=author)
    insert_facts(fact_tuples, author=author)


def _update_facts_from_file(file, author=None):
    fact_values = {}
    with open(file, 'rb') as csvfile:
        reader = rdfreader.get_rdf_reader(csvfile)
        for row in reader:
            if collection.is_empty_list(row):
                continue
            if row[0].startswith('#'):
                continue
            if row[0].startswith('START_TRANSACTION'):
                continue
            if row[0].startswith('COMMIT'):
                _update_facts_from_value_map(fact_values, author=author)
                fact_values = {}
                continue
            if len(row) >= 4:
                fact_values[(row[0], row[1], row[2])] = row[3]
            elif len(row) >= 2:
                fact_values[row[0]] = row[1]


def update_facts(fact_values=None, file=None, author=None):
    if fact_values:
        _update_facts_from_value_map(fact_values, author)
    if file:
        _update_facts_from_file(file, author)


def _erase_facts_by_ids(fact_ids, author=None):
    if eternity_dbutils.is_any_type_facts(fact_ids):
        raise exceptions.InvalidInputValueError(
            'Not allowed to delete any to:type')
    else:
        ers = Eraser(author=author)
        ers.erase_facts(fact_ids)


def _erase_nodes_by_ids(node_ids, author=None):
    ers = Eraser(author=author)
    ers.erase_nodes(node_ids)


def _erase_facts_by_tuples(fact_tuples, author=None):
    fact_ids = eternity_dbutils.get_fact_ids(fact_tuples, author)
    _erase_facts_by_ids(fact_ids, author=author)


def _erase_facts_from_file(file, author=None):
    fact_ids = []
    fact_tuples = []
    with open(file, 'rb') as csvfile:
        reader = rdfreader.get_rdf_reader(csvfile)
        for row in reader:
            if collection.is_empty_list(row):
                continue
            if row[0].startswith('#'):
                continue
            if row[0].startswith('START_TRANSACTION'):
                continue
            if row[0].startswith('COMMIT'):
                if fact_ids:
                    _erase_facts_by_ids(fact_ids, author=author)
                if fact_tuples:
                    _erase_facts_by_tuples(fact_tuples, author=author)
                fact_ids = []
                fact_tuples = []
                continue
            if len(row) >= 3:
                fact_tuples.append((row[0], row[1], row[2]))
            elif len(row) >= 1:
                fact_ids.append(row[0])


def erase_facts(
        fact_ids=None, node_ids=None, fact_tuples=None,
        file=None, author=None):
    if fact_ids:
        _erase_facts_by_ids(fact_ids, author)
    if node_ids:
        _erase_nodes_by_ids(node_ids, author)
    if fact_tuples:
        _erase_facts_by_tuples(fact_tuples, author)
    if file:
        _erase_facts_from_file(file, author)


def graph_query(body):
    ql = GraphQuery()
    return ql.query(body)



import tornado.ioloop
import tornado.web
import tornado.httpserver
import tornado.httputil
import json
import cgi
import datetime


class JSONHandler(tornado.httpserver.HTTPRequest):

    def __call__(self):
        self.stream.read_bytes(self.content_length, self.parse_json)

    def parse_json(self, data):
        print data
        try:
            json_data = json.loads(data)
        except ValueError:
            raise tornado.httpserver._BadRequestException(
                "Invalid JSON structure."
            )
        if type(json_data) != dict:
            raise tornado.httpserver._BadRequestException(
                "We only accept key value objects!"
            )
        for key, value in json_data.iteritems():
            self.request.arguments[key] = [value, ]
        self.done()


class CSRFHandler(tornado.web.RequestHandler):

    def set_default_headers(self):
        self.set_header("Access-Control-Allow-Origin", "*")
        self.set_header("Access-Control-Allow-Headers",
                        "x-requested-with,authorization,content-type")
        self.set_header('Access-Control-Allow-Methods',
                        'POST, GET, OPTIONS, PUT, DELETE')

    def options(self, *args):
        self.set_status(204)
        self.finish()


class XMLMaker(object):

    def __init__(self):
        pass

    def make_xml(self, obj, pretty=False):
        response = '<xml version="1.0" encoding="UTF-8">\n'
        obj = {'response': obj}
        indent = 0
        response += self._sub_make_xml(obj, indent=indent, pretty=pretty)
        return response

    def _sub_make_xml(self, obj, indent, pretty=False):
        def make_entry(k, obj, indent, pretty=False):
            newline = ''
            indent_close = False
            if not pretty:
                indent = 0
            if type(obj) is list or type(obj) is dict:
                subtree = self._sub_make_xml(
                    obj, indent=indent + 2, pretty=pretty)
                newline = '\n'
                indent_close = True
            elif type(obj) is datetime.date:
                if obj >= datetime.datetime(1900, 1, 1).date():
                    subtree = obj.strftime("%Y-%m-%d")
                else:
                    subtree = "%s-%02d-%02d" % (obj.year, obj.month, obj.day)
            elif type(obj) is datetime.datetime:
                if obj >= datetime.datetime(1900, 1, 1):
                    subtree = obj.strftime("%Y-%m-%d")
                else:
                    subtree = "%s-%02d-%02d" % (obj.year, obj.month, obj.day)
            else:
                if type(obj) is unicode:
                    subtree = obj.encode('utf-8')
                else:
                    subtree = str(obj)
            response = ' ' * indent
            response += '<%s>%s%s' % (k, newline, subtree)
            if indent_close:
                response += ' ' * indent
            response += '</%s>\n' % k
            return response

        response = ''
        if type(obj) is dict:
            for k in obj:
                response += make_entry(k, obj[k], indent, pretty)
        elif type(obj) is list:
            for item in obj:
                response += make_entry('item', item, indent, pretty)
        return response


class FloatEncoder(json.JSONEncoder):
    #Todo need docs what it does/when to use/why it is needed

    def __init__(self, nan_str="null", **kwargs):
        super(FloatEncoder, self).__init__(**kwargs)
        self.nan_str = nan_str

    def iterencode(self, o, _one_shot=False):
        """Encode the given object and yield each string
        representation as available.

        For example::

            for chunk in JSONEncoder().iterencode(bigobject):
                mysocket.write(chunk)
        """
        if self.check_circular:
            markers = {}
        else:
            markers = None
        if self.ensure_ascii:
            _encoder = json.encoder.encode_basestring_ascii
        else:
            _encoder = json.encoder.encode_basestring
        if self.encoding != 'utf-8':
            def _encoder(o, _orig_encoder=_encoder, _encoding=self.encoding):
                if isinstance(o, str):
                    o = o.decode(_encoding)
                return _orig_encoder(o)

        def floatstr(o, allow_nan=self.allow_nan, _repr=json.encoder.FLOAT_REPR,
                     _inf=json.encoder.INFINITY, _neginf=-json.encoder.INFINITY,
                     nan_str=self.nan_str):
            # Check for specials.  Note that this type of test is processor
            # and/or platform-specific, so do tests which don't depend on the
            # internals.

            if o != o:
                text = nan_str
            elif o == _inf:
                text = nan_str
            elif o == _neginf:
                text = nan_str
            else:
                return _repr(o)

            if not allow_nan:
                raise ValueError(
                    "Out of range float values are not JSON compliant: " +
                    repr(o))

            return text

        _iterencode = json.encoder._make_iterencode(
            markers, self.default, _encoder, self.indent, floatstr,
            self.key_separator, self.item_separator, self.sort_keys,
            self.skipkeys, _one_shot)
        return _iterencode(o, 0)


class BaseHandler(CSRFHandler):

    JSON = 'json'
    xmlmaker = XMLMaker()

    def _make_json_compatible(self, obj, pretty=False):
        dthandler = lambda obj: (self.jsonize_date(obj) if isinstance(
            obj, datetime.datetime) or isinstance(obj, datetime.date) else None)
        if pretty:
            return json.dumps(obj, default=dthandler, indent=4)
        else:
            return json.dumps(obj, cls=FloatEncoder, default=dthandler)

    def jsonize_date(self, date):
        if (type(date) == datetime.date):
            if (date >= datetime.datetime(1900, 1, 1).date()):
                return date.strftime("%Y-%m-%d")
            else:
                return "%s-%02d-%02d" % (date.year, date.month, date.day)
        elif (type(date) == datetime.datetime):
            if (date >= datetime.datetime(1900, 1, 1)):
                return date.strftime("%Y-%m-%d")
            else:
                return "%s-%02d-%02d" % (date.year, date.month, date.day)
        else:
            return ""

    def apiwrite(self, response, data_type=JSON, status=True):
        '''status is False if an error is thrown'''
        if data_type == self.JSON or data_type == self.XML:
            params = self.request.query
            data = cgi.parse_qs(params)
            format_needed = data.get('format', ['json'])[0]
            pretty = data.get('pretty', ['false'])[0]
            if pretty.lower() == 'true':
                pretty = True
            else:
                pretty = False
            if status:
                final_response = {'status': 'SUCCESS', 'content': response}
            else:
                final_response = {'status': 'FAILURE', 'error': response}
            if format_needed == 'xml':
                self.write(self.xmlmaker.make_xml(final_response, pretty))
            else:
                self.write(self._make_json_compatible(final_response, pretty))




class ToflerDBException(Exception):
    '''super class of any exception raised by toflerdb'''


class RelationAlreadyIsA(ToflerDBException):
    '''return this when there is an attempt to set isA on a relation that is
    already isA something'''


class EntityAlreadyIsA(ToflerDBException):
    '''return this when there is an attempt to set isA on a entity that is
    already that thing'''


class EntityNotFound(ToflerDBException):
    '''return this when an entity was expected to exist in the database
    but didn't '''


class CouldNotSaveError(ToflerDBException):
    '''return this when there was a problem saving an object'''


class MissingRequiredFieldError(ToflerDBException):
    '''return this when any required field is missing'''


class InvalidInputValueError(ToflerDBException):
    '''return this when'''


class WriteLockedNodeError(ToflerDBException):
    '''return this when'''



import logging
import datetime

# import awshelper

FORMAT = '%(asctime)-15s %(levelname)s %(funcName)s() : %(message)s'


class ElasticLogHandler(logging.Handler):
    """This class provides the ability to log directly to elastic search"""

    def __init__(self, elastic):
        logging.Handler.__init__(self)
        self.es = None
        # selfinfo = awshelper.get_instance_details()
        # self.ip = selfinfo['privateIp']
        self.elastic = elastic

    def prepare_record_data(self, record):
        record_data = {
            # 'machine_ip'    : self.ip,
            'created': datetime.datetime.fromtimestamp(record.created),
            'exc_info': record.exc_info,
            'exc_text': record.exc_text,
            'level_name': record.levelname,
            'filename': record.filename,
            'module': record.module,
            'funcname': record.funcName,
            'lineno': record.lineno,
            'msg': record.msg,
            'formatted_msg': self.format(record),
            'process': record.process,
        }
        return record_data

    def put_data(self, record_data, doc_type='superlogger'):
        if self.es is None:
            self.es = self.elastic.get_connection()
        self.es.index(index='logger', doc_type=doc_type, body=record_data)

    def emit(self, record):
        record_data = self.prepare_record_data(record)
        doc_type = 'superlogger'
        if 'audit' in record.args:
            doc_type = 'auditlogger'
            record_data['level_name'] = 'AUDIT'
        self.put_data(record_data, doc_type=doc_type)


class Logger(logging.getLoggerClass()):
    DEBUG = logging.DEBUG
    INFO = logging.INFO
    WARNING = logging.WARNING
    WARN = logging.WARN
    ERROR = logging.ERROR
    CRITICAL = logging.CRITICAL

    def __init__(self, *args, **kwargs):
        super(Logger, self).__init__(*args, **kwargs)
        self.printing_enabled = False

    def enable_filelog(self, filename):
        flh = logging.FileHandler(filename)
        formatter = logging.Formatter(FORMAT)
        flh.setFormatter(formatter)
        self.addHandler(flh)

    def enable_printing(self):
        if self.printing_enabled:
            return
        slh = logging.StreamHandler()
        formatter = logging.Formatter(FORMAT)
        slh.setFormatter(formatter)
        self.addHandler(slh)
        self.printing_enabled = True

    def audit(self, msg, *args, **kwargs):
        self.critical(msg, {'audit': True}, *args, **kwargs)



from toflerdb.utils.common import Common
from toflerdb.utils import collection
from toflerdb.config import SNAPSHOT_INDEX, SNAPSHOT_DOC_TYPE


def insert_entity(entity_id, entity_body):
    es = Common.get_elasticsearch_connection()
    es.index(SNAPSHOT_INDEX, 'entity', body=entity_body, id=entity_id)


def get_root_type(element):
    pass


def get_type(element):
    query = """
        SELECT * FROM toflerdb_ontology WHERE subject = %s AND predicate = %s
    """
    query_data = (element, 'rdfs:type')
    response = Common.execute_query(query, query_data)
    if len(response):
        return response[0]['predicate']
    return None


def get_direct_predicate_value(subj, pred, additional_lookup={}):
    if not subj:
        return []
    if not type(subj) == list:
        subj = [subj]
    or_clauses = ' OR '.join(['subject = %s' for _ in subj])
    query = """
        SELECT subject, predicate, object, value FROM toflerdb_eternity
        WHERE predicate = %s AND
    """
    query += '(' + or_clauses + ')'
    query_data = tuple([pred] + subj)
    # print query
    response = Common.execute_query(query, query_data)
    if len(response) == 0:
        query = """
            SELECT subject, predicate, object, value FROM toflerdb_ontology
            WHERE predicate = %s AND
        """
        query += '(' + or_clauses + ')'
        response = Common.execute_query(query, query_data)

    new_values = []
    for res in response:
        val = res['object']
        if val is None:
            val = res['value']
        new_values.append(val)

    for s in subj:
        if s in additional_lookup and pred in additional_lookup[s]:
            new_values += [x for x in additional_lookup[s]
                           [pred] if x not in new_values]
    return new_values


def get_predicate_value(subj, pred, level=10000, additional_lookup={}):
    # if level >= 0:
    #     root = True
    # if pred == 'rdf:type':
    level -= 1
    new_values = get_direct_predicate_value(subj, pred, additional_lookup)
    if len(new_values) == 0 or level <= 0:
        return new_values

    for item in new_values:
        values = get_predicate_value(item, pred, level=level,
                                     additional_lookup=additional_lookup)
        new_values = collection.merge_unique(new_values, values)
    return new_values


def get_direct_inverse_predicate_value(objc, pred, additional_lookup={}):
    if not objc:
        return []
    if not type(objc) == list:
        objc = [objc]
    or_clauses = ' OR '.join(['object = %s OR value = %s' for _ in objc])
    query = """
        SELECT subject, predicate, object, value FROM toflerdb_eternity
        WHERE predicate = %s AND
    """
    query += '(' + or_clauses + ')'
    objc_data = []
    for x in objc:
        objc_data += [x, x]
    query_data = tuple([pred] + objc_data)
    # print query
    response = Common.execute_query(query, query_data)
    if len(response) == 0:
        query = """
            SELECT subject, predicate, object, value FROM toflerdb_ontology
            WHERE predicate = %s AND
        """
        query += '(' + or_clauses + ')'
        response = Common.execute_query(query, query_data)

    new_values = []
    for res in response:
        val = res['subject']
        new_values.append(val)

    for s in objc:
        if s in additional_lookup and pred in additional_lookup[s]:
            new_values += [x for x in additional_lookup[s]
                           [pred] if x not in new_values]
    return new_values


def get_inverse_predicate_value(objc, pred, level=10000, additional_lookup={}):
    level -= 1
    new_values = get_direct_inverse_predicate_value(
        objc, pred, additional_lookup)
    if len(new_values) == 0 or level <= 0:
        return new_values

    for item in new_values:
        values = get_inverse_predicate_value(
            item, pred, level=level, additional_lookup=additional_lookup)
        new_values = collection.merge_unique(new_values, values)
    return new_values


def exists_in_eternity(subj, additional_lookup={}, eternity_only=False,
                       ontology_only=False):
    response = []
    if not ontology_only:
        query = """
            SELECT subject FROM toflerdb_eternity WHERE subject = %s
        """
        response = Common.execute_query(query, subj)
    if len(response) == 0 and not eternity_only:
        query = """
            SELECT subject FROM toflerdb_ontology WHERE subject = %s
        """
        response = Common.execute_query(query, subj)
    return len(response) > 0 or subj in additional_lookup


def exists_in_ontology(subj, pred, objc):
    query = """
        SELECT subject FROM toflerdb_ontology WHERE subject = %s AND predicate
        = %s AND (object = %s OR value = %s)
    """
    response = Common.execute_query(query, (subj, pred, objc, objc))
    if len(response):
        return True
    return False


def is_unique_predicate_value(pred, additional_lookup={}):
    pred_value = get_predicate_value(
        pred, 'to:isUnique', additional_lookup=additional_lookup)
    is_unique = False
    if len(pred_value) == 1 and pred_value[0]:
        is_unique = True

    return is_unique


def insert_into_ontology(inputs):
    if not inputs:
        return
    query = """
        INSERT INTO toflerdb_ontology(subject, predicate, object, value)
        VALUES
    """
    query_data = []
    query_clauses = []
    for item in inputs:
        query_clauses.append('%s, %s, %s, %s')
        query_data += [item['subject'], item['predicate'],
                       item['object'], item['value']]
    query_clauses_str = '), ('.join(query_clauses)
    query += '(' + query_clauses_str + ')'
    Common.execute_query(query, tuple(query_data))


def get_complete_snapshot_mapping():
    es = Common.get_elasticsearch_connection()
    complete_mapping = es.indices.get_mapping(
        index=SNAPSHOT_INDEX, doc_type=SNAPSHOT_DOC_TYPE)

    return complete_mapping[SNAPSHOT_INDEX]['mappings'][SNAPSHOT_DOC_TYPE]['properties']


def create_snapshot_mapping(mapping):
    es = Common.get_elasticsearch_connection()
    es.indices.put_mapping(index=SNAPSHOT_INDEX,
                           doc_type=SNAPSHOT_DOC_TYPE, body=mapping)


############################################
#          ONTOLOGY ONLY UTILS             #
############################################
def get_all_namespaces():
    query = """
        SELECT DISTINCT(subject) AS subject FROM toflerdb_ontology
    """
    response = Common.execute_query(query)
    all_namespaces = []
    for res in response:
        namespace = res['subject'].split(':')[0]
        if namespace not in all_namespaces:
            all_namespaces.append(namespace)

    return all_namespaces


def get_all_entities_of_namespace(nss):
    if not isinstance(nss, list):
        nss = [nss]
    all_namespaces = get_all_namespaces()
    namespaces = collection.intersection(nss, all_namespaces)
    if not len(namespaces):
        # this is required as the namespace is not a query_data in the
        # following query
        return []
    rx = '|'.join(['%s:' % x for x in namespaces])
    rx = '^(%s).*' % rx
    query = """
        SELECT DISTINCT(subject) AS subject FROM toflerdb_ontology WHERE
        subject REGEXP %s
    """
    response = Common.execute_query(query, rx)
    all_entities = []
    for res in response:
        subj = res['subject']
        subj_subclass = get_predicate_value(subj, 'to:subClassOf')
        if 'to:Entity' in subj_subclass:
            all_entities.append(subj)

    return all_entities


def get_all_properties_where_entity_is_domain_of(entity):
    entity_subclass = get_predicate_value(entity, 'to:subClassOf')
    if not isinstance(entity, list):
        entity = [entity]
    return get_inverse_predicate_value(entity_subclass + entity, 'to:domain')


def get_all_properties_where_entity_is_range_of(entity):
    entity_subclass = get_predicate_value(entity, 'to:subClassOf')
    return get_inverse_predicate_value(entity_subclass + [entity], 'to:range')


def ontology_tree():
    pass


def get_all_propeties(element):
    if not isinstance(element, list):
        element = [element]

    element_subclass = get_predicate_value(element, 'to:subClassOf')
    prop_list = []
    if 'to:RelationalProperty' in element_subclass:
        element_range = get_predicate_value(element, 'to:range')
        element_range_subclass = get_predicate_value(
            element_range, 'to:subClassOf')
        prop_list = get_inverse_predicate_value(
            element_range + element_range_subclass, 'to:domain')
    elif 'to:ComplexRelationalProperty' in element_subclass:
        # if not isinstance(element, list):
        #     element = [element]
        element_range = get_predicate_value(element, 'to:range')
        element_range_subclass = get_predicate_value(
            element_range, 'to:subClassOf')
        element_subclass = get_predicate_value(element, 'to:subClassOf')
        prop_list = get_inverse_predicate_value(
            element_subclass + element + element_range + element_range_subclass, 'to:domain')
    elif 'to:ComplexProperty' in element_subclass or 'to:Entity' in element_subclass:
        # if not isinstance(element, list):
        #     element = [element]
        # element_subclass = get_predicate_value(element, 'to:subClassOf')
        prop_list = get_inverse_predicate_value(
            element_subclass + element, 'to:domain')

    ret_val = []
    for prop in prop_list:
        ret_val.append({
            'label': prop,
            'type': get_predicate_value(prop, 'to:subClassOf'),
            'ontolabel': get_predicate_value(prop, 'to:ontoLabel'),
            'ontodesc': get_predicate_value(prop, 'to:ontoDescription'),
        })

    return ret_val


def get_all_propeties_tree(element, retval={}):
    '''
        this method will return a tree structure as opposed to the method
        get_all_propeties. The output structure would be
        {
            <prop_name>: {
                ontolabel: '',
                ontodesc: '',
                range: '',
                children: {
                    <child_prop_name>: {
                        ontolabel: '',
                        ontodesc: '',
                        range: '',
                        children: ...
                    }
                }
            },

        }
    '''
    if not isinstance(element, list):
        element = [element]

    element_subclass = get_predicate_value(element, 'to:subClassOf')
    if collection.intersection(element_subclass, ['to:ComplexProperty', 'to:Entity']):
        pass

    pass



from toflerdb.utils.common import Common


def is_locked(node):
    r = Common.get_cache_connection()
    return r.exists(node)


def lock_nodes(nodes):
    LOCK_INTERVAL = 5 * 60
    if not isinstance(nodes, list):
        nodes = [nodes]
    r = Common.get_cache_connection()
    for node in nodes:
        r.set(node, node, ex=LOCK_INTERVAL, nx=True)
    return nodes


def release_lock(nodes):
    r = Common.get_cache_connection()
    r.delete(nodes)



'''MySQL easy-access module

This module makes access to a MySQL database very simple.
It manages the connection pools, etc, converts the returned rows into
dictionaries

For update/delete queries, it returns information about
the number of rows affected.

For insert queries, it returns the lastrowid in case the
affected table contains an auto increment column
'''

import mysql.connector

IntegrityError = mysql.connector.errors.IntegrityError


class MySQLCursorDict(mysql.connector.cursor.MySQLCursor):
    '''This is a utility class that converts the rows returned from a MySQL
    query into a list of dictionaries where each column becomes a key'''

    def _row_to_python(self, rowdata, desc=None):
        row = super(MySQLCursorDict, self)._row_to_python(rowdata, desc)
        if row:
            return dict(zip(self.column_names, row))
        return None


class Database(object):

    def __init__(self, config, pool_name='toflerdb'):
        self._config = config
        self._pool_name = pool_name

    def execute_query(self, query, params=None, **kwargs):
        """Execute a specified query on the MySQL database.

        Args:
            query (str): The query to execute
            params (tuple): List of parameter values to match the number of
                placeholder items in the query

        Returns:
            list/dict: For select queries, returns a list of dictionaries
                       For update/delete/insert queries, returns a dictionary
                       containing *rows* (count of affected rows) and
                       *lastrowid* which contains value of any
                       auto_increment field in case of an insert
        """

        if params is not None and not isinstance(params, tuple) and not isinstance(params, dict):
            params = (params,)
        conn = mysql.connector.connect(
            pool_name=self._pool_name, **self._config)
        cursor = conn.cursor(cursor_class=MySQLCursorDict)
        try:
            cursor.execute(query, params, **kwargs)
        except Exception as e:
            conn.close()
            raise e
        try:
            # if an insert or update query was called, an exception will
            # result. In that case, we return the count of inserted/updated
            # rows and the lastrowid in case there is an auto-increment field
            results = cursor.fetchall()
            conn.commit()
        except:
            conn.commit()
            results = {'rows': cursor.rowcount, 'lastrowid': cursor.lastrowid}
        conn.close()
        return results



# import tornado.web
import json
from toflerdb.core import api as gcc_api
from basehandler import BaseHandler


class UnauthorizedQuery(ValueError):
    '''Returned when query tries to access a resource not allowed for a user'''


class QueryHandler(BaseHandler):

    def post(self):
        return self.get()

    def get(self):
        request_body = self.request.body
        if request_body is None:
            return []
        response = []
        try:
            request_body = json.loads(request_body)
        except:
            print "Error processing request"
            response = []
        try:
            response = self.process_request(request_body)
            self.apiwrite(response)
        except UnauthorizedQuery:
            # TODO: find a way to specify the exact unauthorized fields
            self.apiwrite('Unauthorized Query', status=False)

    def screen_query(self, query):
        # All fields specified in the query must be
        # permitted for the user. otherwise, the query
        # will be rejected

        # Step 1 is to check for namespaces. All public
        # namespaces go through by default

        # Step 2 - for private namespaces, individual
        # types and properties are to be checked based
        # on dataset permissions. This can be cached
        # on a per-user basis?

        if 'to:type' in query:
            types = query['to:type']
            if type(types) is not list:
                types = list[types]

        elif 'id' not in query:
            query['to:type'] = 'to:Entity'

        return True

    def filter_responses(self, responses):
        return responses

    def process_request(self, request):
        #valid_query = self.screen_query(request)
        # if not valid_query:
        #    raise UnauthorizedQuery("Query to access non-permitted fields")
        '''
        if_event_type = False
        if 'to:type' in request:
            if request['to:type'] == 'gobiz:Investment'
                    or request['to:type'] == 'gomarkets:IndiaBulkDeal':
                if_event_type = True
        '''
        output = gcc_api.graph_query(request)
        filtered_output = self.filter_responses(output)
        return filtered_output



import json
import traceback
# from toflerdb.utils.common import Common
from toflerdb.core import api as gcc_api
from toflerdb.utils import exceptions
from basehandler import BaseHandler


class UploadFactsHandler(BaseHandler):

    def post(self):
        request_body = self.request.body
        if request_body is None:
            return []
        response = []
        try:
            request_body = json.loads(request_body)
        except:
            print "Error processing request"
            response = []
        fact_tuples = request_body.get('fact_tuples', None)
        file_text = request_body.get('file_text', None)
        ignore_duplicate = request_body.get('ignore_duplicate', True)
        author = request_body.get('author', None)
        try:
            response = gcc_api.insert_facts(
                fact_tuples=fact_tuples, file_text=file_text, author=author,
                ignore_duplicate=ignore_duplicate)
            self.apiwrite(response)
        except exceptions.ToflerDBException, e:
            print traceback.format_exc()
            self.apiwrite(str(e), status=False)
        except Exception, e:
            print traceback.format_exc()
            # Common.get_logger().error(str(e))
            self.apiwrite('Something went wrong', status=False)



def is_templatized_id(temp_id):
    try:
        retvalue = temp_id.startswith('_')
    except:
        return False
    return retvalue


def is_keep_id(temp_id):
    return is_templatized_id(temp_id) and not temp_id.startswith('__')


def append_userid(temp_id, uid):
    if not uid:
        raise Exception(
            'Use of templatized id without user id is strictly restricted')

    return '%s#%s' % (temp_id, uid)


def trim_userid(temp_ids, uid):
    temp_id_map = {}
    end_str = '#%s' % uid
    for i in temp_ids:
        val = temp_ids[i]
        if is_templatized_id(i) and i.endswith(end_str):
            temp_id_map[i.split(end_str)[0]] = val

    return temp_id_map



from toflerdb.utils import exceptions
from toflerdb.utils.common import Common
from toflerdb.dbutils import dbutils
from toflerdb.utils import collection


class OntoMaker(object):

    def __init__(self):
        self._sql_query_list = []
        self._elastic_query_list = []
        self._entity_id = None
        self._entity_type_list = None
        self._input_list = []
        self._normalized_input = {}
        self._inverse_normalized_input = {}
        self._stores = {
            'snapshot': {},
            'inmemory': {}
        }
        self._new_mapping = {}
        self._complete_mapping = dbutils.get_complete_snapshot_mapping()
        self._validation = True

    def turn_off_validation(self):
        self._validation = False
        return self

    def turn_on_validation(self):
        self._validation = True
        return self

    def normalize_input(self, subj, pred, objc):
        if subj not in self._normalized_input:
            self._normalized_input[subj] = {}
        if pred not in self._normalized_input[subj]:
            self._normalized_input[subj][pred] = []
        if objc not in self._normalized_input[subj][pred]:
            self._normalized_input[subj][pred].append(objc)

    def inverse_normalize_input(self, subj, pred, objc):
        if subj not in self._inverse_normalized_input:
            self._inverse_normalized_input[subj] = {}
        if pred not in self._inverse_normalized_input[subj]:
            self._inverse_normalized_input[subj][pred] = []
        if objc not in self._inverse_normalized_input[subj][pred]:
            self._inverse_normalized_input[subj][pred].append(objc)

    def delete_normalize_input(self, subj, pred, objc):
        self._normalized_input[subj][pred].remove(objc)

    def add_input(self, subj, pred, objc, prev=None, author=None):
        self.normalize_input(subj, pred, objc)
        self.inverse_normalize_input(objc, pred, subj)
        if not dbutils.exists_in_ontology(subj, pred, objc):
            if self._validation:
                self.validate(subj, pred, objc)

            self._input_list.append({
                'subject': subj,
                'predicate': pred,
                'object': objc,
                'fact_id': Common.generate_id(None)
            })
        return self

    def validate(self, subj, pred, objc):
        '''
        if pred == to:subClassOf:
            (objc + objc_subclass) in [to:Entity, to:Property, to:Literal]
        elif pred == to:ontoLabel:
            subj must be define i.e. subj_subclass is non empty
        elif pred == to:description:
            subj must be define i.e. subj_subclass is non empty
        elif pred == to:domain:
            objc needs to exists in db as subject
        elif pred == to:range:
            objc needs to exists in db as subject
            if subj_subclass is Relational/Complex Relational Property:
                subj_range_subclass in [to:Entity, to:Property]
            elif subj_subclass is Property:
                subj_range_subclass is [to:Literal]
        elif pred == 'to:isUnique':
            pass
        else:
            not valid
        '''
        input_tuple = (subj, pred, objc)
        (subj_subclass, ) = self.gather_info_from_toflerdb(
            subj, ['subclass'])
        if not subj_subclass:
            error_txt = (
                "Subject is not defined : %s"
                "\nInput tuple : %s") % (subj, input_tuple)
            Common.get_logger().error(error_txt)
            raise exceptions.InvalidInputValueError(error_txt)

        if pred == 'to:subClassOf':
            (objc_subclass, ) = self.gather_info_from_toflerdb(
                objc, ['subclass'])
            if not collection.intersection(
                    objc_subclass + [objc],
                    ['to:Entity', 'to:Property', 'to:Literal']):
                error_txt = (
                    "Superclass is not defined : %s"
                    "\nInput tuple : %s") % (objc, input_tuple)
                Common.get_logger().error(error_txt)
                raise exceptions.InvalidInputValueError(error_txt)

        elif pred == 'to:ontoLabel':
            pass
        elif pred == 'to:description':
            pass
        elif pred == 'to:domain':
            if not dbutils.exists_in_eternity(
                    objc, additional_lookup=self._normalized_input,
                    ontology_only=True):
                error_txt = (
                    "Domain is not defined : %s"
                    "\nInput tuple : %s") % (objc, input_tuple)
                Common.get_logger().error(error_txt)
                raise exceptions.InvalidInputValueError(error_txt)
        elif pred == 'to:range':
            if not dbutils.exists_in_eternity(
                    objc, additional_lookup=self._normalized_input,
                    ontology_only=True):
                error_txt = (
                    "Range is not defined : %s"
                    "Input tuple : %s") % (objc, input_tuple)
                Common.get_logger().error(error_txt)
                raise exceptions.InvalidInputValueError(error_txt)

            (objc_subclass, ) = self.gather_info_from_toflerdb(
                objc, ['subclass'])
            if collection.intersection(
                    [
                        'to:RelationalProperty',
                        'to:ComplexRelationalProperty'
                    ], subj_subclass):
                if not collection.intersection(
                        [objc] + objc_subclass,
                        ['to:Entity', 'to:Property']):
                    error_txt = (
                        "For relational property declaration, range must"
                        "be either of type to:Entity or to:Property."
                        "\nProperty : %s, range type: %s."
                        "\nInput tuple : %s"
                    ) % (subj, objc_subclass, input_tuple)
                    Common.get_logger().error(error_txt)
                    raise exceptions.InvalidInputValueError(error_txt)
            elif 'to:Property' in subj_subclass:
                if not 'to:Literal' in objc_subclass:
                    error_txt = (
                        "For property declaration,"
                        "range must be of type to:Literal."
                        "\nProperty : %s, range type: %s."
                        "\nInput tuple : %s"
                    ) % (subj, objc_subclass, input_tuple)
                    Common.get_logger().error(error_txt)
                    raise exceptions.InvalidInputValueError(error_txt)
        elif pred == 'to:isUnique':
            pass
        else:
            error_txt = (
                "Predicate must be one of to:subClassOf, to:ontoLabel,"
                " to:description, to:domain, to:domain, to:range."
                "\nInput Tuple : %s") % str(input_tuple)
            Common.get_logger().error(error_txt)
            raise exceptions.InvalidInputValueError(error_txt)

    def gather_info_from_toflerdb(self, elem, prop_list):
        ret_val = []
        if elem not in self._normalized_input:
            self._normalized_input[elem] = {}
        for prop in prop_list:
            if not prop in self._normalized_input[elem]:
                if prop == 'type':
                    self._normalized_input[elem][prop] = \
                        dbutils.get_predicate_value(
                            elem, 'to:type',
                            additional_lookup=self._normalized_input)
                elif prop == 'subclass':
                    self._normalized_input[elem][prop] = \
                        dbutils.get_predicate_value(
                            elem, 'to:subClassOf',
                            additional_lookup=self._normalized_input)
                elif prop == 'domain':
                    self._normalized_input[elem][prop] = \
                        dbutils.get_predicate_value(
                            elem, 'to:domain', level=1,
                            additional_lookup=self._normalized_input)
                elif prop == 'range':
                    self._normalized_input[elem][prop] = \
                        dbutils.get_predicate_value(
                            elem, 'to:range',
                            additional_lookup=self._normalized_input)
                elif prop == 'type_subclass':
                    (prop_type, ) = self.gather_info_from_toflerdb(
                        elem, ['type'])
                    self._normalized_input[elem][prop] = \
                        dbutils.get_predicate_value(
                            prop_type, 'to:subClassOf',
                            additional_lookup=self._normalized_input)
                elif prop == 'type_domain':
                    (prop_type, ) = self.gather_info_from_toflerdb(
                        elem, ['type'])
                    self._normalized_input[elem][prop] = \
                        dbutils.get_predicate_value(
                            prop_type, 'to:domain', level=1,
                            additional_lookup=self._normalized_input)
                elif prop == 'type_range':
                    (prop_type, ) = self.gather_info_from_toflerdb(
                        elem, ['type'])
                    self._normalized_input[elem][prop] = \
                        dbutils.get_predicate_value(
                            prop_type, 'to:range',
                            additional_lookup=self._normalized_input)
                elif prop == 'domain_subclass':
                    (prop_domain, ) = self.gather_info_from_toflerdb(
                        elem, ['domain'])
                    self._normalized_input[elem][prop] = \
                        dbutils.get_predicate_value(
                            prop_domain, 'to:subClassOf',
                            additional_lookup=self._normalized_input)
                elif prop == 'range_subclass':
                    (prop_range, ) = self.gather_info_from_toflerdb(
                        elem, ['range'])
                    self._normalized_input[elem][prop] = \
                        dbutils.get_predicate_value(
                            prop_range, 'to:subClassOf',
                            additional_lookup=self._normalized_input)

            ret_val.append(self._normalized_input[elem][prop])

        return tuple(ret_val)

    def make_mapping(self):
        all_subjects = []
        for info in self._input_list:
            if info['subject'] not in all_subjects:
                all_subjects.append(info['subject'])

        for subj in all_subjects:
            # subj = info['subject']
            # pred = info['predicate']
            # objc = info['object']
            (
                subj_subclass,
                subj_domain,
                subj_domain_subclass,
                subj_range,
                subj_range_subclass
            ) = self.gather_info_from_toflerdb(subj, [
                'subclass', 'domain', 'domain_subclass',
                'range', 'range_subclass'])
            # print 'subj : %s\nsubj_subclass : %s\nsubj_domain :
            # %s\nsubj_domain_subclass : %s\nsubj_range :
            # %s\nsubj_range_subclass : %s' %(subj, subj_subclass, subj_domain,
            # subj_domain_subclass, subj_range, subj_range_subclass)
            if 'to:Property' not in subj_subclass:
                continue
            if 'to:Property' in subj_domain_subclass:
                for dmn in subj_domain:
                    path = collection.find_path(self._complete_mapping, dmn)
                    if path is None:
                        continue
                    path.append('properties')
                    value = {
                        subj: {
                            'properties': {
                                'fact_id': {
                                    'type': 'string',
                                    'index': 'not_analyzed'
                                },
                            }
                        }
                    }
                    if collection.intersection(
                            [
                                'to:ComplexProperty',
                                'to:ComplexRelationalProperty'
                            ], subj_subclass):
                        value[subj]['type'] = 'nested'
                    
                    if 'to:Null' in subj_range_subclass:
                        # here we don't want any value key
                        pass
                    elif collection.intersection(
                            [
                                'to:RelationalProperty',
                                'to:ComplexRelationalProperty'
                            ], subj_subclass):
                        # the value would be a reference to another node
                        value[subj]['properties']['value'] = {
                            'type': 'string', 'index': 'not_analyzed'}
                    else:
                        value[subj]['properties']['value'] = \
                            collection.get_datatype(
                                subj_range + subj_range_subclass)
                    collection.assign_value_to_path(
                        self._complete_mapping, value, path)
                    collection.assign_value_to_path(
                        self._new_mapping, value, path)
            elif 'to:Entity' in subj_domain + subj_domain_subclass:
                value = {
                    subj: {
                        'properties': {
                            'fact_id': {
                                'type': 'string',
                                'index': 'not_analyzed'
                            },
                        }
                    }
                }
                if collection.intersection(
                        [
                            'to:ComplexProperty',
                            'to:ComplexRelationalProperty'
                        ], subj_subclass):
                    value[subj]['type'] = 'nested'

                if 'to:Null' in subj_range_subclass:
                    # here we don't want any value key
                    pass
                elif collection.intersection(
                        [
                            'to:RelationalProperty',
                            'to:ComplexRelationalProperty'
                        ], subj_subclass):
                    # the value would be a reference to another node
                    value[subj]['properties']['value'] = {
                        'type': 'string',
                        'index': 'not_analyzed'
                    }
                else:
                    value[subj]['properties']['value'] = \
                        collection.get_datatype(
                            subj_range + subj_range_subclass)
                self._complete_mapping.update(value)
                self._new_mapping.update(value)
            # add all the superclass properties also
            if len(subj_subclass):
                subj_subclass_domain_of = dbutils.get_inverse_predicate_value(
                    subj_subclass, 'to:domain', level=1,
                    additional_lookup=self._inverse_normalized_input)
                # print 'subj_subclass_domain_of : %s\n\n'
                # %subj_subclass_domain_of
                for ssdo in subj_subclass_domain_of:
                    path = collection.find_path(self._complete_mapping, subj)
                    if not path:
                        continue
                    path.append('properties')
                    (ssdo_range, ) = self.gather_info_from_toflerdb(
                        ssdo, ['range'])
                    value = {
                        ssdo: {
                            'properties': {
                                'fact_id': {
                                    'type': 'string',
                                    'index': 'not_analyzed'
                                },
                                'value': collection.get_datatype(ssdo_range)
                            }
                        }
                    }
                    collection.assign_value_to_path(
                        self._new_mapping, value, path)
                    collection.assign_value_to_path(
                        self._complete_mapping, value, path)

    def make_ontology_inputs(self):
        for info in self._input_list:
            if not dbutils.exists_in_eternity(info['object']):
                info['value'] = info['object']
                info['object'] = None
            else:
                info['value'] = None

    def commit(self):
        self.make_ontology_inputs()
        dbutils.insert_into_ontology(self._input_list)
        if self._new_mapping:
            dbutils.create_snapshot_mapping({'properties': self._new_mapping})



'''GraphQuery

This module provides the core capabilities to process GraphQuery queries
to run on top of ToflerDB
'''

import json
from pydash import merge, get
from toflerdb.dbutils import dbutils
from toflerdb.dbutils import snapshot as snapshot_dbutils
from toflerdb.utils import collection


class GraphQuery(object):

    def __init__(self):
        self._predicate_values = {}
        self._to_delete = False
        self.agg_operators = ['count', 'avg', 'sum', 'groupBy', 'distinctCount', 'min', 'max']

    def _gather_info_from_toflerdb(self, elem, prop_list):
        (elem, op) = self._extract_operator(elem)
        ret_val = []
        if elem not in self._predicate_values:
            self._predicate_values[elem] = {}
        for prop in prop_list:
            if not prop in self._predicate_values[elem]:
                if prop == 'type':
                    self._predicate_values[elem][prop] = \
                        dbutils.get_predicate_value(elem, 'to:type')
                elif prop == 'subclass':
                    self._predicate_values[elem][prop] = \
                        dbutils.get_predicate_value(elem, 'to:subClassOf')
                elif prop == 'domain':
                    self._predicate_values[elem][prop] = \
                        dbutils.get_predicate_value(elem, 'to:domain', level=1)
                elif prop == 'range':
                    self._predicate_values[elem][prop] = \
                        dbutils.get_predicate_value(elem, 'to:range')
                elif prop == 'type_subclass':
                    (prop_type, ) = self._gather_info_from_toflerdb(
                        elem, ['type'])
                    self._predicate_values[elem][prop] = \
                        dbutils.get_predicate_value(prop_type, 'to:subClassOf')
                elif prop == 'domain_subclass':
                    (prop_domain, ) = self._gather_info_from_toflerdb(
                        elem, ['domain'])
                    self._predicate_values[elem][prop] = \
                        dbutils.get_predicate_value(prop_domain, 'to:subClassOf')
                elif prop == 'type_domain':
                    (prop_type, ) = self._gather_info_from_toflerdb(
                        elem, ['type'])
                    self._predicate_values[elem][prop] = \
                        dbutils.get_predicate_value(
                            prop_type, 'to:domain', level=1)
                elif prop == 'type_range':
                    (prop_type, ) = self._gather_info_from_toflerdb(
                        elem, ['type'])
                    self._predicate_values[elem][prop] = \
                        dbutils.get_predicate_value(prop_type, 'to:range')
                elif prop == 'subclass_inverse_domain':
                    (prop_subclass, ) = self._gather_info_from_toflerdb(
                        elem, ['subclass'])
                    self._predicate_values[elem][prop] = \
                        dbutils.get_inverse_predicate_value(
                            [elem] + prop_subclass, 'to:domain', level=1)
            ret_val.append(self._predicate_values[elem][prop])

        return tuple(ret_val)

    def _has_graph_filter(self, request):
        for f in request:
            if isinstance(request[f], dict):
                if self._has_graph_filter(request[f]):
                    return True
            elif request[f] is not None:
                return True
        return False

    def _prepare_one_hop_request_body(self, elem, request):
        (elem_inverse_domain, ) = self._gather_info_from_toflerdb(
            elem, ['subclass_inverse_domain'])
        newrequest = {}
        for f in request:
            (f_ex, op) = self._extract_operator(f)
            if f_ex not in elem_inverse_domain:
                newrequest[f] = request[f]

        return newrequest

    def _prepare_direct_request_body(self, elem, request):
        (elem_inverse_domain, ) = self._gather_info_from_toflerdb(
            elem, ['subclass_inverse_domain'])
        newrequest = {}
        for f in request:
            (f_ex, op) = self._extract_operator(f)
            if f_ex in elem_inverse_domain:
                newrequest[f] = request[f]

        return newrequest

    def _extract_operator(self, f):
        all_operators = [
            '~|=', '|=', '~=', '>', '>=', '<',
            '<=', '~[]=', '~[]', '[]=', '[]'] + self.agg_operators

        op = None
        fld = f
        for o in all_operators:
            if f.endswith(o):
                op = o
                fld = f.split(o)[0].strip()
                break

            if f.startswith(o):
                op = o
                fld = f.rsplit("(")[1].strip().rsplit(")", 1)[0].strip()
                break
        return (fld, op)

    def _create_leaf_match_query(self, fld, fld_val):
        ''' this method checks for to:label field and create match
        query accordingly
        '''
        leaf_match_query = None
        if fld == 'to:label.value':
            leaf_match_query = {
                'bool': {
                    'should': [
                        {"match": {fld: fld_val}},
                        {"match": {"%s.value" % fld: fld_val}},
                        {"match": {"%s.whitespace_edgegram" % fld: fld_val}},
                        {"match": {"%s.whitespace_keyword" % fld: fld_val}}
                    ],
                    'minimum_should_match': 1
                }
            }
        else:
            leaf_match_query = {'match': {fld: fld_val}}
        return leaf_match_query

    def _operator_query_builder(self, op, fld, fld_val, org_fld, base_fld):
        op_query = None
        is_agg_query = is_nested_query = False
        if op is None:
            op_query = self._create_leaf_match_query(fld, fld_val)
        elif op == '~|=':
            op_query = {'bool': {'must_not': []}}
            if not isinstance(fld_val, list):
                fld_val = [fld_val]
            for fv in fld_val:
                op_query['bool']['must_not'].append(
                    self._create_leaf_match_query(fld, fv))
        elif op == '~=':
            op_query = {
                'bool': {
                    'must_not': [self._create_leaf_match_query(fld, fld_val)]
                }
            }
        elif op == '|=':
            op_query = {
                'bool': {
                    'should': [],
                    'minimum_should_match': 1
                }
            }
            if not isinstance(fld_val, list):
                fld_val = [fld_val]
            for fv in fld_val:
                op_query['bool']['should'].append(
                    self._create_leaf_match_query(fld, fv))
        elif op == '>=':
            op_query = {
                'range': {
                    fld: {
                        'gte': fld_val
                    }
                }
            }
        elif op == '>':
            op_query = {
                'range': {
                    fld: {
                        'gt': fld_val
                    }
                }
            }
        elif op == '<=':
            op_query = {
                'range': {
                    fld: {
                        'lte': fld_val
                    }
                }
            }
        elif op == '<':
            op_query = {
                'range': {
                    fld: {
                        'lt': fld_val
                    }
                }
            }
        elif op == '~[]=':
            op_query = {
                'bool': {
                    'should': [
                        {'range': {fld: {'lt': fld_val[0]}}},
                        {'range': {fld: {'gt': fld_val[1]}}}
                    ],
                    'minimum_should_match': 1
                }
            }
        elif op == '[]=':
            op_query = {
                'range': {
                    fld: {
                        'gte': fld_val[0],
                        'lte': fld_val[1]
                    }
                }
            }
        elif op == '~[]':
            op_query = {
                'bool': {
                    'should': [
                        {'range': {fld: {'lte': fld_val[0]}}},
                        {'range': {fld: {'gte': fld_val[1]}}}
                    ],
                    'minimum_should_match': 1
                }
            }
        elif op == '[]':
            op_query = {
                'range': {
                    fld: {
                        'gt': fld_val[0],
                        'lt': fld_val[1]
                    }
                }
            }
        elif op in self.agg_operators:
            is_agg_query = True
            if op == 'groupBy':
                op_query = {
                    fld: {
                        "terms": {
                            "field": fld
                        }
                    }
                }
            elif op == 'count':
                op_query = {
                    fld: {
                        "value_count": {
                            "field": fld
                            # "precision_threshold": 100,
                            # "rehash": false
                        }
                    }
                }
            elif op == 'distinctCount':
                op_query = {
                    fld: {
                        "cardinality": {
                            "field": fld
                            # "precision_threshold": 100,
                            # "rehash": false
                        }
                    }
                }
            elif op == 'avg':
                op_query = {
                    fld: {
                        "avg": {
                            "field": fld
                        }
                    }
                }
            elif op == 'sum':
                op_query = {
                    fld: {
                        "sum": {
                            "field": fld
                        }
                    }
                }
            elif op == 'min':
                op_query = {
                    fld: {
                        "top_hits": {
                            "size": 1,
                            "_source": {
                                "includes": [
                                    base_fld + "id",
                                    fld
                                ]
                            },
                            "sort": {
                                fld: "asc"
                            }
                        }
                    },
                }
            elif op == 'max':
                op_query = {
                    fld: {
                        "top_hits": {
                            "size": 1,
                            "_source": {
                                "includes": [
                                    base_fld + "id",
                                    fld
                                ]
                            },
                            "sort": {
                                fld: "desc"
                            }
                        }
                    },
                }

        (fld_domain_subclass,) = self._gather_info_from_toflerdb(org_fld, ['domain_subclass'])
        # print "base fld", fld, op, org_fld, fld_domain_subclass, base_fld
        if collection.intersection(fld_domain_subclass, [
                'to:ComplexRelationalProperty', 'to:ComplexProperty']):
            is_nested_query = True
            if base_fld.endswith('.'):
                base_fld = base_fld[:-1]
            if is_agg_query:
                op_query = {
                    base_fld+"-nested": {
                        "nested": {
                            "path": base_fld
                        },
                        "aggs": op_query
                    }
                }
            else:
                op_query = {
                    "nested": {
                        "path": base_fld.split(".")[-1],
                        "query": op_query,
                        "inner_hits": {}
                    }
                }

        return is_agg_query, is_nested_query, op_query

    def _get_field_name(self, f, base_string):

        if f == 'id':
            field_name = base_string + f
        elif f == 'value':
            field_name = base_string + f
        else:
            field_name = base_string + f + ".value"
        return field_name

    def _create_match_array(self, request, match_array, agg_query, base_string=""):
        nested_queries = {}
        # is_agg_query_scope = any([key.startswith(
        #     op) for op in self.agg_operators for key in request])
        for f in request:
            f_val = request[f]
            (f, op) = self._extract_operator(f)
            if f_val is None and op is None:
                continue

            if type(f_val) is dict:
                newrequest = f_val
                (f_subclass, ) = self._gather_info_from_toflerdb(
                    f, ['subclass'])
                if collection.intersection(f_subclass, [
                        'to:ComplexRelationalProperty',
                        'to:RelationalProperty']):
                    newrequest = self._prepare_direct_request_body(
                        f, f_val)

                new_agg_query = {}
                child_nested_queries = self._create_match_array(
                    newrequest, match_array, new_agg_query, base_string + f + ".")

                if new_agg_query:
                    agg_query.setdefault("aggs", {}).update(new_agg_query)

                if child_nested_queries:
                    grp_name = base_string.replace(".", "_") if base_string else "root"
                    agg_query.setdefault(grp_name, {
                        "terms": {"field": "id"},
                        "aggs": {}
                    })["aggs"].update(child_nested_queries)

            else:
                field_name = self._get_field_name(f, base_string)
                is_agg_query, is_nested_query, op_query = self._operator_query_builder(
                    op, field_name, f_val, f, base_string)
                if not is_agg_query:
                    match_array.append(op_query)
                else:
                    #for merging if agg belongs to same nested scope
                    if is_nested_query:
                        merge(nested_queries, op_query)
                    else:
                        agg_query.update(op_query)

        return nested_queries

    def _key_in_aggdata(self, key, aggdata, refval):
        for k in aggdata:
            if k == key:
                refval[0] = aggdata[k]
                return True
            elif isinstance(aggdata[k], dict) and self._key_in_aggdata(key, aggdata[k], refval):
                return True
            elif isinstance(aggdata[k], list):
                for item in aggdata[k]:
                    if isinstance(item, dict) and self._key_in_aggdata(key, item, refval):
                        return True
        return None

    def _create_aggregation_output(self, request, aggdata, base_string):
        op = {}
        has_agg_req_keys = False

        for agg_req_key in request.keys():
            (f, operator) = self._extract_operator(agg_req_key)
            field_name = self._get_field_name(f, base_string)
            refval = [None]
            if operator in self.agg_operators:
                has_agg_req_keys = True
                if aggdata and self._key_in_aggdata(field_name, aggdata, refval):
                    agg_value = refval[0]

                    # if :
                    #     #list
                    # else:
                        #not list

                    if operator in ['min', 'max']:
                        agg_value = get(refval[0], "hits.hits[0]._source.%s" % (f), None)
                        if agg_value:
                            op.update({"id": get(refval[0], "hits.hits[0]._source.id", None)})
                    op[agg_req_key] = agg_value

        return has_agg_req_keys, op

    def _create_output(self, data, request, output, aggdata, base_string=""):
        if self._to_delete:
            return

        for f in request:
            f_val = request[f]
            (f, op) = self._extract_operator(f)
            if type(f_val) is dict:
                if f not in data:
                    if self._has_graph_filter(f_val):
                        self._to_delete = True
                        return
                    continue
                op = None
                '''
                if f is a subclassof of either to:ComplexRelationalProperty
                or to:RelationalProperty,
                query all the ids in within itself and replace those inplace
                with associated
                '''
                (f_subclass, ) = self._gather_info_from_toflerdb(
                    f, ['subclass'])
                if collection.intersection(f_subclass, [
                        'to:ComplexRelationalProperty',
                        'to:RelationalProperty']):
                    if type(data[f]) is list:
                        #multiple object return case
                        newdata = []
                        hop_ids = []
                        hop_ids = [item['value'] for item in data[f]
                                   if item['value'] not in hop_ids]
                        filter_id = f_val.get('id', None)
                        newrequest = None
                        if filter_id and filter_id in hop_ids:
                            #fetch filtered id
                            newrequest = self._prepare_one_hop_request_body(
                                f, f_val)
                            newrequest.update({'id': filter_id})
                        elif not filter_id:
                            #fetch all ids
                            newrequest = self._prepare_one_hop_request_body(
                                f, f_val)
                            newrequest.update({'id|=': hop_ids})

                        if newrequest:
                            newresponse, newresponse_aggdata = self.query(
                                newrequest, raw_output=True)
                            if len(newresponse):
                                newresponse_normalized = {}
                                for nr in newresponse:
                                    newresponse_normalized[nr['id']] = nr
                                for item in data[f]:
                                    if item['value'] in newresponse_normalized:#checking id in newresponse_normalized
                                        item.update(newresponse_normalized[
                                            item['value']])
                                        newdata.append(item)

                        if len(newdata):
                            data[f] = newdata
                        else:
                            self._to_delete = True
                            return
                    else:
                        #single object return case
                        filter_id = f_val.get('id', None)
                        hop_id = data[f]['value']
                        #ignore unmatched ones
                        if filter_id and hop_id != filter_id:
                            continue
                        newrequest = self._prepare_one_hop_request_body(
                            f, f_val)
                        newrequest.update({'id': hop_id})
                        newresponse, newresponse_aggdata = self.query(
                            newrequest, raw_output=True)
                        if len(newresponse):
                            data[f].update(newresponse[0])
                        else:
                            self._to_delete = True
                            return

                    has_agg_req_keys, op = self._create_aggregation_output(f_val, newresponse_aggdata, "")
                else:
                    #aggdata so replacing all with aggdata
                    has_agg_req_keys, op = self._create_aggregation_output(f_val, aggdata, base_string+f+".")

                if not has_agg_req_keys:
                    if type(data[f]) is list:
                        # print 'list when val is dict', f
                        op = []
                        for item in data[f]:
                            if 'fact_id' in item:
                                item.pop('fact_id')
                            new_op = {}
                            self._create_output(item, f_val, new_op, aggdata, base_string + f + ".")
                            op.append(new_op)
                    else:
                        op = {}
                        data[f].pop('fact_id')
                        # print 'val when val is dict', f
                        self._create_output(data[f], f_val, op, base_string + f + ".")
                output[f] = op

            else:

                if f in data:
                    if type(data[f]) is list:
                        # print 'list when val is not dict', f
                        new_array = []
                        for item in data[f]:
                            if f == 'value':
                                new_array.append(item)
                            else:
                                new_array.append(item['value'])
                        output[f] = new_array
                    else:
                        # print 'val when val is not dict', f
                        if f == 'id' or f == 'value':
                            output[f] = data[f]
                        else:
                            output[f] = data[f]['value']
                else:
                    print f, 'not in data when val is not dict'

    def query(self, request, raw_output=False):
        match_array = []
        agg_query = {}
        output = []
        self._create_match_array(request, match_array, agg_query)
        elastic_query = {
            "query": {
                "bool": {
                    "must": match_array
                }
            }
        }
        if agg_query:
            elastic_query["aggs"] = agg_query

        print "req", request, elastic_query
        try:
            response = snapshot_dbutils.execute_search_query(
                elastic_query, size=10)
            aggout = response.get("aggregations")
            print "aggregations", aggout, len(response['hits']['hits'])
        except Exception as e:
            print str(e)

        has_agg_req_keys, op = self._create_aggregation_output(request, aggout, "")
        if has_agg_req_keys and not raw_output:
            output.append(op)
        else:
            hits = response['hits']['hits']
            for hit in hits:
                self._to_delete = False
                src = hit['_source']
                if 'inner_hits' in hit:
                    for nested_path in hit['inner_hits']:
                        src[nested_path] = [
                            nested_hit['_source']
                            for nested_hit in
                            hit['inner_hits'][nested_path]['hits']['hits']]
                opsrc = {}
                if not raw_output:
                    ## pass aggout
                    self._create_output(src, request, opsrc, aggout)
                    if not self._to_delete:
                        output.append(opsrc)
                else:
                    output.append(src)

        if not raw_output:
            return output
        else:
            return output, aggout




import json
from toflerdb.utils.common import Common


def insert_task(user_id, public_taskid, internal_taskid, task_queue_id,
                task_status, task_info):
    query = """
        INSERT INTO toflerdb_bulk_task_queue(user_id, public_taskid,
        internal_taskid, task_queue_id, task_status, task_info)
        VALUES (%s, %s, %s, %s, %s, %s)
    """
    query_data = (user_id, public_taskid, internal_taskid, task_queue_id,
                  task_status, json.dumps(task_info))
    Common.execute_query(query, query_data)


def get_task_by_id(taskid):
    query = """
        SELECT user_id, public_taskid, internal_taskid, task_queue_id,
        task_status, task_info FROM toflerdb_bulk_task_queue WHERE
        public_taskid = %s OR internal_taskid = %s
    """
    query_data = (taskid, taskid)
    response = Common.execute_query(query, query_data)
    if len(response) == 1:
        retval = response[0]
        retval['task_info'] = json.loads(retval['task_info'])
        return retval
    return {}


def update_task_status(public_taskid, new_status):
    query = """
        UPDATE toflerdb_bulk_task_queue SET task_status = %s
        WHERE public_taskid = %s
    """
    query_data = (new_status, public_taskid)
    Common.execute_query(query, query_data)


def update_task_info(public_taskid, new_info):
    query = """
        UPDATE toflerdb_bulk_task_queue SET task_info = %s
        WHERE public_taskid = %s
    """
    query_data = (json.dumps(new_info), public_taskid)
    Common.execute_query(query, query_data)


def get_tasks_of_user(user_id, task_queue_id):
    query = """
        SELECT user_id, public_taskid, internal_taskid, task_queue_id,
        task_status, task_info FROM toflerdb_bulk_task_queue WHERE
        user_id = %s AND task_queue_id = %s
    """
    query_data = (user_id, task_queue_id)

    return Common.execute_query(query, query_data)



from toflerdb.dbutils import eraser as eraser_dbutils
from toflerdb.utils import collection


class Eraser(object):

    def __init__(self, **kwargs):
        self._author = kwargs.get('author')

    def populate_subfacts(self, fact_ids):
        subfact_ids = eraser_dbutils.find_subfact_ids(fact_ids)
        return collection.merge_unique(fact_ids, subfact_ids)

    def erase_facts(self, fact_ids):
        # type deletion must be strictly restricted
        if not isinstance(fact_ids, list):
            fact_ids = [fact_ids]
        # we need to delete all the associated facts about any complex
        # ( realtional) fact from eternity also.
        # if the given fact id is a fact id of a complex (realtional/)property
        # add all the sub fact ids of that node to the deletion list
        fact_ids = self.populate_subfacts(fact_ids)
        eraser_dbutils.erase_facts_from_eternity(fact_ids, author=self._author)
        eraser_dbutils.erase_facts_from_snapshot(fact_ids)

    def erase_nodes(self, node_ids):
        if not isinstance(node_ids, list):
            node_ids = [node_ids]

        fact_ids = eraser_dbutils.find_related_facts_by_node_id(node_ids)
        self.erase_facts(fact_ids)



import time


class IDManager(object):

    _MAX_WORKERS = 1024 * 8
    _LEASE_PERIOD = 60  # seconds
    _EVENTS_PER_SECOND = 400000

    ALPHABET = 'bcdfghjklmnpqrstvwxyz0123456789'
    ALPHABET_REVERSE = dict((c, i) for (i, c) in enumerate(ALPHABET))
    BASE = len(ALPHABET)

    def __init__(self, db, logger):
        self._counter = 0
        self._counter_reset_at = 0
        self._WORKER_ID = None
        self._db_lease_ts = None
        self._obtained_at = None
        self._db = db
        self._logger = logger
        self._obtain_new_worker_id()

    def _obtain_new_worker_id(self):
        query = """DELETE FROM toflerdb_worker WHERE
                assigned_at < NOW() - INTERVAL lease_period SECOND
                """
        self._db.execute_query(query)

        query = """SELECT worker_id FROM toflerdb_worker ORDER BY
                worker_id"""
        block_query = """INSERT INTO toflerdb_worker (worker_id) VALUES (%s)"""
        lease_start_tsquery = """SELECT assigned_at FROM toflerdb_worker
                    WHERE worker_id = %s"""
        candidate = 1
        while True:
            response = self._db.execute_query(query)
            if len(response) == self._MAX_WORKERS:
                self._logger.warn(
                    'Worker pool saturated. Sleeping for 1 second.')
                time.sleep(1)
            for res in response:
                if res['worker_id'] == candidate:
                    candidate += 1
                else:
                    break
            if candidate > self._MAX_WORKERS:
                self._logger.warn(
                    'Worker pool saturated. Sleeping for 1 second.')
                time.sleep(1)
                continue
            try:
                self._db.execute_query(block_query, candidate)
                self._WORKER_ID = candidate
                tsquery = self._db.execute_query(
                    lease_start_tsquery, candidate)
                self._db_lease_ts = tsquery[0]['assigned_at']
                self._obtained_at = time.time()
                break
            except:
                pass

    def _renew_lease(self):
        query = """UPDATE toflerdb_worker SET assigned_at = NOW()
                WHERE worker_id = %s AND assigned_at = %s"""
        response = self._db.execute_query(
            query, (self._WORKER_ID, self._db_lease_ts))
        if response['rows'] == 0:
            self._logger.warn(
                "Lease expired before renewal. Obtaining new lease.")
            self._obtain_new_worker_id()
            return
        self._obtained_at = time.time()
        query = """SELECT assigned_at FROM toflerdb_worker WHERE worker_id = %s"""
        response = self._db.execute_query(query, self._WORKER_ID)
        self._db_lease_ts = response[0]['assigned_at']
        self._logger.debug("Lease renewed for worker id %s" % self._WORKER_ID)

    def generate_new_id(self, namespace=None):
        now = time.time()
        if self._obtained_at + self._LEASE_PERIOD < now:
            self._renew_lease()
        timestamp = int(now * 1000)

        '''the following this is an interesting trick. We want to reset the
        counter every 1 ms. So, we multiply now by 1000 and take the floor.
        If that changes, we reset'''
        if timestamp > self._counter_reset_at:
            self._counter_reset_at = timestamp
            self._counter = 1
        if self._counter >= 4090:
            '''TODO: for now the counter is fine. But ensure that
            it resets by the next call'''
            time.sleep(1)   # TODO: 1s is too long. We want it to be 2-3 ms.
        candidate = (timestamp * self._MAX_WORKERS + self._WORKER_ID) * 4096 + self._counter
        self._counter += 1
        if namespace is None:
            return self.num_encode(candidate)
        return 't:%s/%s' % (namespace, self.num_encode(candidate))

    def num_encode(self, n):
        s = []
        while True:
            n, r = divmod(n, self.BASE)
            s.append(self.ALPHABET[r])
            if n == 0:
                break
        return ''.join(reversed(s))



import json
import traceback
from toflerdb.core import api as gcc_api
from toflerdb.utils import exceptions
from basehandler import BaseHandler


class UploadOntologyHandler(BaseHandler):

    def post(self):
        request_body = self.request.body
        if request_body is None:
            return []
        response = []
        try:
            request_body = json.loads(request_body)
        except:
            print "Error processing request"
            response = []
        onto_tuples = request_body.get('onto_tuples', None)
        onto_text = request_body.get('file_text', None)
        try:
            response = gcc_api.insert_ontology(
                onto_tuples=onto_tuples, file_text=onto_text)
            self.apiwrite(response)
        except exceptions.ToflerDBException, e:
            print traceback.format_exc()
            self.apiwrite(str(e), status=False)
        except:
            print traceback.format_exc()
            self.apiwrite('Something went wrong', status=False)



import csv


class RDF(csv.Dialect):
    delimiter = ' '
    quotechar = '"'
    # escapechar = '"'
    lineterminator = '\n'
    quoting = csv.QUOTE_MINIMAL

csv.register_dialect('rdfttl', RDF)


def get_rdf_reader(file_handle):
    reader = csv.reader(file_handle, 'rdfttl')
    return reader


def get_rdf_writer(file_handle):
    writer = csv.writer(file_handle, 'rdfttl')
    return writer



# import json
import datetime
from toflerdb.utils.common import Common
from toflerdb.utils import collection
from toflerdb.config import FACT_STATUS
import snapshot as snapshot_dbutils


def erase_facts_from_eternity(fact_ids, author=None):
    if not isinstance(fact_ids, list):
        fact_ids = [fact_ids]

    query = """
        UPDATE toflerdb_eternity SET status = %s, status_updated_on = %s,
        status_updated_by = %s WHERE fact_id IN
    """
    placeholder_str = ', '.join(['%s' for _ in fact_ids])
    query = '%s (%s)' % (query, placeholder_str)
    query_data = tuple([FACT_STATUS.DELETED, datetime.datetime.now(), author] +
                       fact_ids)
    Common.execute_query(query, query_data)


def is_empty_node(node):
    if node is None:
        return True
    if len(node.keys()) == 1 and node.keys()[0] == 'id':
        return True

    return False


def erase_facts_from_snapshot(fact_ids):
    if not isinstance(fact_ids, list):
        fact_ids = [fact_ids]

    # get all the entities where factid matches
    # for each entity delete specific segment of the object
    # reinsert all the entities into elastic

    matched_entities = snapshot_dbutils.get_snapshot_nodes(fact_id=fact_ids)
    processed_entities = None
    missing_entities = None
    if matched_entities:
        # print json.dumps(matched_entities, indent=4)
        processed_entities = collection.delete_segment(
            matched_entities, list(map(lambda fid: {'fact_id': fid},
                                       fact_ids)))
        # print json.dumps(processed_entities, indent=4)
    if processed_entities:
        snapshot_dbutils.insert_into_snapshot(processed_entities)
        missing_entities = [entity_id for entity_id in matched_entities
                            if is_empty_node(
                                processed_entities.get(entity_id))]

    if missing_entities:
        snapshot_dbutils.delete_snapshot_nodes_by_id(missing_entities)


def find_subfact_ids(fact_ids):
    if not isinstance(fact_ids, list):
        fact_ids = [fact_ids]

    query = """
        SELECT predicate FROM toflerdb_eternity WHERE fact_id IN
    """
    placeholder_str = ', '.join(['%s' for _ in fact_ids])
    query = '%s (%s)' % (query, placeholder_str)
    query_data = tuple(fact_ids)
    response = Common.execute_query(query, query_data)
    pred = [res['predicate'] for res in response]
    query = """
        SELECT fact_id FROM toflerdb_eternity WHERE subject IN
    """
    placeholder_str = ', '.join(['%s' for _ in pred])
    query = '%s (%s)' % (query, placeholder_str)
    query_data = tuple(pred)
    response = Common.execute_query(query, query_data)
    retval = [res['fact_id'] for res in response]

    return retval


def find_related_facts_by_node_id(node_ids):
    if not isinstance(node_ids, list):
        node_ids = [node_ids]

    placeholder_str = ', '.join(['%s' for _ in node_ids])
    query = """
        SELECT fact_id FROM toflerdb_eternity WHERE subject IN (%s)
        OR object IN (%s)
    """ % (placeholder_str, placeholder_str)
    query_data = tuple(node_ids + node_ids)
    response = Common.execute_query(query, query_data)

    return [res['fact_id'] for res in response]



import hashlib


def merge_unique(arr1, arr2):
    return arr1 + [val for val in arr2 if val not in arr1]


def intersection(arr1, arr2):
    return [val for val in arr2 if val in arr1]


def subtract(arr1, arr2):
    return [val for val in arr1 if val not in arr2]


def is_sublist(arr, sub_arr):
    for e in sub_arr:
        if e not in arr:
            return False
    return True


def is_equal_list(arr1, arr2):
    return is_sublist(arr1, arr2) and is_sublist(arr2, arr1)


def is_empty_list(arr):
    if not arr:
        return True
    for elem in arr:
        if isinstance(elem, str) or isinstance(elem, unicode):
            elem = elem.strip()
        if elem:
            return False

    return True

# def find_path(element, JSON, path, all_paths):
#   if element in JSON:
#     path = path + element + ' = ' + JSON[element].encode('utf-8')
#     print path
#     all_paths.append(path)
#   for key in JSON:
#     if isinstance(JSON[key], dict):
#       find(element, JSON[key],path + key + '.',all_paths)


def add_fact_id(val, fid, is_unique=True):
    obj = {
        'fact_id': fid,
        'value': val
    }
    return obj if is_unique else [obj]


def found_id(JSON, _id):
    if isinstance(JSON, list):
        for j in JSON:
            if isinstance(j, dict) and j.get('id') == _id:
                return True
    elif isinstance(JSON, dict) and JSON.get('id') == _id:
        return True
    return False


def is_subsegment(JSON, segment):
    if not isinstance(segment, list):
        segment = [segment]

    for s in segment:
        if s.viewitems() <= JSON.viewitems():
            return True
    return False


def delete_segment(JSON, segment):
    """
    This method deletes the segment of the JSON object where the
    argument segment is a sub match in the JSON. It also clears out
    any key with empty elem.

    Example1:
        Input:
            JSON = {
                "key1" : "val1",
                "key2" : {
                    "key2.1" : "val2.1"
                }
            }
            segment = {"key2.1" : "val2.1"}
        Output:
            JSON = {
                "key1" : "val1",
            }

    Example2:
        Input:
            JSON = {
                "key1" : "val1",
                "key2" : [{
                    "key2_1.1" : "val2_1.1"
                }, {
                    "key2_2.1" : "val2_2.1"
                }]
            }
            segment = {"key2_1.1" : "val2_1.1"}
        Output:
            JSON = {
                "key1" : "val1",
                "key2" : [{
                    "key2_2.1" : "val2_2.1"
                }]
            }

    Example3:
        Input:
            JSON = {
                "key1" : "val1",
                "key2" : [{
                    "key2_1.1" : "val2_1.1"
                }, {
                    "key2_2.1" : "val2_2.1"
                }, {
                    "key2_3.1" : "val2_3.1",
                    "key2_3.2" : "val2_3.2"
                }, {
                    "key2_4.1" : "val2_4.1"
                }]
            }
            segment = [
                {"key2_1.1" : "val2_1.1"},
                {"key2_3.2" : "val2_3.2"}
            ]
        Output:
            JSON = {
                "key1" : "val1",
                "key2" : [{
                    "key2_2.1" : "val2_2.1"
                }, {
                    "key2_4.1" : "val2_4.1"
                }]
            }
    """
    if isinstance(JSON, dict) and is_subsegment(JSON, segment):
        return None
    elif isinstance(JSON, list):
        to_del_idx = []
        for idx in range(0, len(JSON)):
            if delete_segment(JSON[idx], segment) is None:
                to_del_idx.append(idx)

        idx_shift = 0
        for idx in to_del_idx:
            del JSON[idx - idx_shift]
            idx_shift += 1

    elif isinstance(JSON, dict):
        for key in JSON.keys():
            val = JSON[key]
            if delete_segment(val, segment) is None or \
                    (isinstance(val, list) and len(val) == 0):
                del JSON[key]

    return JSON


def find_path(JSON, key, _id=None):
    if isinstance(JSON, dict) and key in JSON and (
            _id is None or found_id(JSON[key], _id)):
        return [key]
    if isinstance(JSON, list):
        for j in JSON:
            return find_path(j, key, _id)
    if isinstance(JSON, dict):
        for k in JSON:
            path = find_path(JSON[k], key, _id)
            if path:
                return [k] + path


def assign_value_to_path(JSON, val, path):
    curr_json = JSON
    for key in path:
        if key not in curr_json:
            curr_json[key] = {}
        curr_json = curr_json[key]
    curr_json.update(val)


def _update_with_value(JSON, value):
    for k in value:
        v = value[k]
        if isinstance(v, list) and k in JSON:
            JSON[k] += v
        else:
            JSON[k] = v


def _direct_assign(JSON, value, _id=None):
    if isinstance(JSON, list):
        for j in JSON:
            if isinstance(j, dict):
                if _id is None:
                    _update_with_value(j, value)
                elif j.get('id') == _id:
                    _update_with_value(j, value)
    if isinstance(JSON, dict):
        if _id is None:
            _update_with_value(JSON, value)
        elif JSON.get('id') == _id:
            _update_with_value(JSON, value)


def assign_value(JSON, key, value, _id=None):
    if not isinstance(value, dict):
        raise ValueError('Value must be a dictionary')
    if isinstance(JSON, dict) and key in JSON:
        _direct_assign(JSON[key], value, _id)

    if isinstance(JSON, list):
        for j in JSON:
            assign_value(j, key, value, _id)

    if isinstance(JSON, dict):
        for k in JSON:
            assign_value(JSON[k], key, value, _id)


def get_datatype(ontology_types):
    if 'to:Date' in ontology_types:
        return {"type": "date"}
    if 'to:Integer' in ontology_types:
        return {"type": "integer"}
    if 'to:Float' in ontology_types:
        return {"type": "float"}
    if 'to:String' in ontology_types:
        return {"type": "string"}
    if 'to:Token' in ontology_types:
        return {"type": "string", "index": "not_analyzed"}
    return {"type": "string"}


def typecast(elem, ontology_types):
    if 'to:Date' in ontology_types:
        return elem
    if 'to:Integer' in ontology_types:
        return int(elem)
    if 'to:Float' in ontology_types:
        return float(elem)
    if 'to:String' in ontology_types:
        return str(elem)
    if 'to:Token' in ontology_types:
        return str(elem)
    return elem


def adjust_start_commit_text(text):
    if not text.startswith('START_TRANSACTION'):
        text = '%s\n%s' % ('START_TRANSACTION .', text)
    if not (text.endswith('COMMIT') or text.endswith('COMMIT .')):
        text = '%s\n%s' % (text, 'COMMIT .')

    return text


def generate_hash(text):
    return hashlib.sha256(text).hexdigest()


def nested_set(dic, keys, value):
    for key in keys[:-1]:
        dic = dic.setdefault(key, {})
    dic[keys[-1]] = value



import time
from toflerdb.utils.common import Common
from toflerdb.utils import exceptions
from toflerdb.dbutils import dbutils
from toflerdb.dbutils import snapshot as snapshot_dbutils
from toflerdb.dbutils import eternity as eternity_dbutils
from toflerdb.utils import collection
from toflerdb.utils import templatizedid
from toflerdb.utils import locks


class EntryMaker(object):

    def __init__(self, **kwargs):
        self._sql_query_list = []
        self._elastic_query_list = []
        self._entity_id = None
        self._entity_type_list = None
        self._input_list = []
        self._normalized_input = {}
        self._stores = {
            'snapshot': {},
            'inmemory': {}
        }
        self._fact_tuple = None
        self._templatized_id_map = {}
        self._new_templatized_id = []
        self._validation = True
        self._fact_id = None
        self._author = kwargs.get('author')
        self._locked_nodes = []
        self._ignore_duplicate = kwargs.get('ignore_duplicate', True)

    def turn_off_validation(self):
        self._validation = False
        return self

    def turn_on_validation(self):
        self._validation = True
        return self

    def normalize_input(self, subj, pred, objc):
        if subj not in self._normalized_input:
            self._normalized_input[subj] = {}
        if pred not in self._normalized_input[subj]:
            self._normalized_input[subj][pred] = []
        if objc not in self._normalized_input[subj][pred]:
            self._normalized_input[subj][pred].append(objc)

    def delete_normalize_input(self, subj, pred, objc):
        self._normalized_input[subj][pred].remove(objc)

    def add_fact_id(self, val, is_unique, is_type=False):
        obj = {
            'fact_id': self._fact_id,
            'value': val if not is_type else self.add_superclasses(val)
        }
        return obj if is_unique else [obj]

    def typecast_input(self, subj, pred, objc):
        # subj = int(subj)
        if pred == 'to:type':
            return objc
        # if not dbutils.exists_in_eternity(pred, ontology_only = True):
        #     pred = int(pred)
        (pred_subclass, pred_type_subclass, pred_range, pred_type_range) = \
            self.gather_info_from_toflerdb(pred, [
                'subclass', 'type_subclass', 'range', 'type_range'])
        if collection.intersection(
                ['to:RelationalProperty', 'to:ComplexRelationalProperty'],
                pred_subclass + pred_type_subclass):
            objc = str(objc)
        else:
            objc = collection.typecast(objc, pred_range + pred_type_range)
        return objc

    def add_templatized_id_map(self, temp_id):
        '''
        create the templatized_id to tofler_id map, if not exists
        returns the tofler_id
        if the temp_id is not a throw away id, look for it in eternity
        if found, cache the mapping, else create a new id
        '''
        if temp_id not in self._templatized_id_map:
            stored_id = []
            if templatizedid.is_keep_id(temp_id):
                stored_id = dbutils.get_inverse_predicate_value(
                    temp_id, 'to:templatizedId', level=1)
            if len(stored_id):
                tofler_id = stored_id[0]
                self._templatized_id_map[temp_id] = tofler_id
            else:
                tofler_id = Common.generate_id(None)
                self._templatized_id_map[temp_id] = tofler_id
                self._new_templatized_id.append(temp_id)
        return self._templatized_id_map[temp_id]

    def handle_templatized_id(self, elem, additional_check=None,
                              additional_check_args=None):
        '''
        given a templatized id the method will return tofler_id
        after caching the tofler_id to templatized_id map for future reference
        templatized_id rule:
            if a id starts with _(underscore), it is a templatized_id
            if a id starts with __(double underscore), it is a throw away id
        '''

        if templatizedid.is_templatized_id(elem):
            if additional_check is None or \
                    additional_check(
                        elem, additional_check_args=additional_check_args):
                temp_id = templatizedid.append_userid(elem, self._author)
                tofler_id = self.add_templatized_id_map(temp_id)
                elem = tofler_id

        return elem

    def is_templatized_object(self, elem, additional_check_args):
        pred = additional_check_args[1]
        if pred == 'to:templatizedId':
            return False
        (pred_subclass, pred_type_subclass) = self.gather_info_from_toflerdb(
            pred, ['subclass', 'type_subclass'])
        if collection.intersection(
                ['to:RelationalProperty', 'to:ComplexRelationalProperty'],
                pred_subclass + pred_type_subclass):
            return True
        return False

    def add_input(self, subj, pred, objc, prev=None):
        self._fact_tuple = (subj, pred, objc)
        subj = self.handle_templatized_id(subj)
        pred = self.handle_templatized_id(pred)
        objc = self.handle_templatized_id(
            objc, additional_check=self.is_templatized_object,
            additional_check_args=(subj, pred, objc))
        if self._ignore_duplicate and \
                eternity_dbutils.tuple_exists_in_eternity(subj, pred, objc):
            return self

        self.normalize_input(subj, pred, objc)
        if self._validation and not self.is_valid(subj, pred, objc):
            self.delete_normalize_input(subj, pred, objc)
            return self
        self._fact_id = Common.generate_id(None)

        self.create_eternity_input(subj, pred, objc, prev)
        self.create_snapshot_input(subj, pred, objc)
        return self

    def gather_info_from_toflerdb(self, elem, prop_list):
        ret_val = []
        if elem not in self._normalized_input:
            self._normalized_input[elem] = {}
        for prop in prop_list:
            if not prop in self._normalized_input[elem]:
                if prop == 'type':
                    self._normalized_input[elem][prop] = \
                        dbutils.get_predicate_value(
                            elem, 'to:type',
                            additional_lookup=self._normalized_input)
                elif prop == 'subclass':
                    self._normalized_input[elem][prop] = \
                        dbutils.get_predicate_value(
                            elem, 'to:subClassOf',
                            additional_lookup=self._normalized_input)
                elif prop == 'domain':
                    self._normalized_input[elem][prop] = \
                        dbutils.get_predicate_value(
                            elem, 'to:domain', level=1,
                            additional_lookup=self._normalized_input)
                elif prop == 'range':
                    self._normalized_input[elem][prop] = \
                        dbutils.get_predicate_value(
                            elem, 'to:range',
                            additional_lookup=self._normalized_input)
                elif prop == 'type_subclass':
                    (prop_type, ) = self.gather_info_from_toflerdb(
                        elem, ['type'])
                    self._normalized_input[elem][prop] = \
                        dbutils.get_predicate_value(
                            prop_type, 'to:subClassOf',
                            additional_lookup=self._normalized_input)
                elif prop == 'type_domain':
                    (prop_type, ) = self.gather_info_from_toflerdb(
                        elem, ['type'])
                    self._normalized_input[elem][prop] = \
                        dbutils.get_predicate_value(
                            prop_type, 'to:domain', level=1,
                            additional_lookup=self._normalized_input)
                elif prop == 'type_range':
                    (prop_type, ) = self.gather_info_from_toflerdb(
                        elem, ['type'])
                    self._normalized_input[elem][prop] = \
                        dbutils.get_predicate_value(
                            prop_type, 'to:range',
                            additional_lookup=self._normalized_input)
            ret_val.append(self._normalized_input[elem][prop])

        return tuple(ret_val)

    def is_valid(self, subj, pred, objc):
        subj_type, subj_type_subclass = self.gather_info_from_toflerdb(
            subj, ['type', 'type_subclass'])
        pred_type, pred_subclass, pred_type_subclass = \
            self.gather_info_from_toflerdb(
                pred, ['type', 'subclass', 'type_subclass'])

        # if subj_type must not be empty
        if not subj_type:
            error_txt = (
                'Subject type must be defined : %s\nInput Tuple : %s'
                % (subj, self._fact_tuple))
            # Common.get_logger().error(error_txt)
            raise exceptions.InvalidInputValueError(error_txt)
            return False

        if not 'to:Property' in pred_type_subclass + pred_subclass:
            error_txt = (
                'Predicate root type must be of to:Property : %s'
                '\nInput Tuple : %s' % (pred, self._fact_tuple))
            # Common.get_logger().error(error_txt)
            raise exceptions.InvalidInputValueError(error_txt)
            return False

        # if pred is 'to:type', we need to special handle the case
        # check whether the object exists in ontology
        if pred == 'to:type' and not dbutils.exists_in_eternity(
                objc, ontology_only=True):
            error_txt = (
                'The type does not exist in ontology: %s\nInput Tuple : %s'
                % (objc, self._fact_tuple))
            # Common.get_logger().error(error_txt)
            raise exceptions.InvalidInputValueError(error_txt)
            return False

        # if pred isA to:RelationalProperty, conditions to satisfy
        # 1. object needs to exists in toflerdb
        # 2. pred_domain should be in subj_type + subject_type_subclass
        # 3. pred_range should be in objc_type + objc_type_subclass
        elif 'to:RelationalProperty' in pred_subclass:
            if not dbutils.exists_in_eternity(
                    objc, additional_lookup=self._normalized_input):
                error_txt = (
                    'Object does not exists in toflerdb : %s\nInput Tuple : %s'
                    % (objc, self._fact_tuple))
                # Common.get_logger().error(error_txt)
                raise exceptions.InvalidInputValueError(error_txt)
                return False
            (pred_domain, pred_range) = self.gather_info_from_toflerdb(
                pred, ['domain', 'range'])
            if not collection.intersection(
                    pred_domain, subj_type + subj_type_subclass):
                error_txt = (
                    'Predicate domain does not satisfy ontology : %s'
                    '\nInput Tuple : %s' % (pred, self._fact_tuple))
                # Common.get_logger().error(error_txt)
                raise exceptions.InvalidInputValueError(error_txt)
                return False
            (objc_type, objc_type_subclass) = self.gather_info_from_toflerdb(
                objc, ['type', 'type_subclass'])
            if not collection.intersection(
                    pred_range, objc_type + objc_type_subclass):
                error_txt = (
                    'Predicate range does not satisfy ontology : %s'
                    '\nInput Tuple : %s' % (pred, self._fact_tuple))
                # Common.get_logger().error(error_txt)
                raise exceptions.InvalidInputValueError(error_txt)
                return False

        # if pred isA to:ComplexRelationalProperty, conditions to satisfy
        # 1. object needs to exists in toflerdb
        # 2. pred_type_domain should be in subj_type + subject_type_subclass
        # 3. pred_type_range should be in objc_type + objc_type_subclass
        elif 'to:ComplexRelationalProperty' in pred_type_subclass:
            if not dbutils.exists_in_eternity(
                    objc, additional_lookup=self._normalized_input):
                error_txt = (
                    'Object does not exists in toflerdb : %s\nInput Tuple : %s'
                    % (objc, self._fact_tuple))
                # Common.get_logger().error(error_txt)
                raise exceptions.InvalidInputValueError(error_txt)
                return False
            (pred_type_domain, pred_type_range) = \
                self.gather_info_from_toflerdb(
                    pred, ['type_domain', 'type_range'])
            if not collection.intersection(
                    pred_type_domain, subj_type + subj_type_subclass):
                error_txt = (
                    'Predicate domain does not satisfy ontology : %s'
                    '\nInput Tuple : %s' % (pred, self._fact_tuple))
                # Common.get_logger().error(error_txt)
                raise exceptions.InvalidInputValueError(error_txt)
                return False
            (objc_type, objc_type_subclass) = self.gather_info_from_toflerdb(
                objc, ['type', 'type_subclass'])
            if not collection.intersection(
                    pred_type_range, objc_type + objc_type_subclass):
                error_txt = (
                    'Predicate range does not satisfy ontology : %s'
                    '\nInput Tuple : %s' % (pred, self._fact_tuple))
                # Common.get_logger().error(error_txt)
                raise exceptions.InvalidInputValueError(error_txt)
                return False

        # if pred isA to:ComplexProperty, conditions to satisfy
        # 1. pred_type_domain should be in subj_type + subject_type_subclass
        elif 'to:ComplexProperty' in pred_type_subclass:
            (pred_type_domain, ) = self.gather_info_from_toflerdb(
                pred, ['type_domain'])
            if not collection.intersection(
                    pred_type_domain, subj_type + subj_type_subclass):
                error_txt = (
                    'Predicate domain does not satisfy ontology : %s'
                    '\nInput Tuple : %s' % (pred, self._fact_tuple))
                # Common.get_logger().error(error_txt)
                raise exceptions.InvalidInputValueError(error_txt)
                return False

        # if pred isA to:Property, conditions to satisfy
        # 1. pred_domain should be in subj_type + subject_type_subclass
        else:
            (pred_domain, ) = self.gather_info_from_toflerdb(pred, ['domain'])
            if not collection.intersection(
                    pred_domain, subj_type + subj_type_subclass):
                error_txt = (
                    'Predicate domain does not satisfy ontology : %s'
                    '\nInput Tuple : %s' % (pred, self._fact_tuple))
                # Common.get_logger().error(error_txt)
                raise exceptions.InvalidInputValueError(error_txt)
                return False

        return True

    def create_eternity_input(self, subj, pred, objc, prev=None):
        eternity_value = None
        (pred_type, pred_subclass, pred_type_subclass) = \
            self.gather_info_from_toflerdb(
                pred, ['type', 'subclass', 'type_subclass'])
        if 'to:ComplexRelationalProperty' in pred_type_subclass or \
                'to:RelationalProperty' in pred_subclass:
            # the object is reference to some other entity
            pass
        else:
            eternity_value = objc
            objc = None
        (subj, pred, objc) = eternity_dbutils.convert_elements_to_string(
            subj, pred, objc)
        (subj, pred, eternity_value) = \
            eternity_dbutils.convert_elements_to_string(
                subj, pred, eternity_value)
        self._input_list.append({
            'subject': subj,
            'predicate': pred,
            'object': objc,
            'value': eternity_value,
            'fact_id': self._fact_id,
            'author': self._author,
            'prev': prev
        })

    def get_node_from_snapshot(self, subj, pred=None, objc=None):
        if pred == 'to:type':
            return
        if locks.is_locked(subj) and subj not in self._locked_nodes:
            raise exceptions.WriteLockedNodeError(
                'The node is locked for writing: %s\nInput Tuple : %s'
                % (subj, self._fact_tuple))
        (subj_type, ) = self.gather_info_from_toflerdb(subj, ['type'])
        if not collection.find_path(self._stores, subj) and \
                not collection.find_path(self._stores, subj_type[0], _id=subj):
            attempts = 2
            while attempts > 0:
                attempts -= 1
                node = snapshot_dbutils.get_snapshot_nodes(id=subj)
                if not node:
                    # Common.get_logger().error(
                    # 'Invalid id, ATTEMPTS Again : %s, %s' % (subj, pred))
                    time.sleep(1)
                else:
                    self._stores['snapshot'].update(node)
                    self._locked_nodes += locks.lock_nodes(
                        [key for key in node])
                    break

    def add_superclasses(self, elem):
        return self.gather_info_from_toflerdb(elem, ['subclass'])[0] + [elem]

    def create_snapshot_input(self, subj, pred, objc):
        (subj_type, subj_type_subclass) = self.gather_info_from_toflerdb(
            subj, ['type', 'type_subclass'])
        (pred_type, pred_type_subclass, pred_type_range) = \
            self.gather_info_from_toflerdb(
                pred, ['type', 'type_subclass', 'type_range'])

        # if pred_type_subclass isA to:ComplexProperty or
        # to:ComplexRelationalProperty, this means the complex property is
        # getting assigned to some other node or some other complex property
        # we will be assigning value key only if pred_type_range is not
        # 'to:Null'.
        # if the subj exists in _stores, this means subject is either entity in which case
        # the subj would be found in _stores['snapshot'] or
        # the subj is an instance of any ComplexRelationalProperty/ ComplexProperty,
        # in which case the subject is found in _stores['memory']
        # if the subj does not exists in _stores as key, this means the subject
        # is an instance of ComplexRelationalProperty/ ComplexProperty, in which case
        # the subject can be found under the subj_type key with subject as id
        # we need to assign the value {pred_type : objc} at the appropriate
        # place
        if collection.intersection(
                ['to:ComplexProperty', 'to:ComplexRelationalProperty'],
                pred_type_subclass):
            self.get_node_from_snapshot(subj, pred)
            if not pred in self._stores['inmemory']:
                raise exceptions.InvalidInputValueError(
                    'ComplexProperty property instance can be used only once :'
                    ' %s\nInput Tuple : %s' % (pred, self._fact_tuple))

            value = self._stores['inmemory'][pred]
            is_unique = dbutils.is_unique_predicate_value(pred_type[0])
            value.update({
                'fact_id': self._fact_id
            })
            if 'to:Null' not in pred_type_range:
                value.update({
                    'value': self.typecast_input(subj, pred, objc)
                    if 'to:type' not in pred_type
                    else self.add_superclasses(objc)
                })
            if not is_unique:
                value = [value]
            if collection.find_path(self._stores, subj):
                collection.assign_value(self._stores, subj, {
                                        pred_type[0]: value})
            else:
                collection.assign_value(self._stores, subj_type[0], {
                                        pred_type[0]: value}, _id=subj)
            del self._stores['inmemory'][pred]

        # if subj_type_subclass isA to:ComplexProperty or to:ComplexRelationalProperty
        # subject is either a key in _stores['memory'] or id value of key subj_type
        # find the subject as key or find subj_type with subject as fact_id
        # assign {pred : objc} under proper key
        elif collection.intersection(
                ['to:ComplexProperty', 'to:ComplexRelationalProperty'],
                subj_type_subclass):
            # this means we have got a complex relation.
            # get the object from elasticsearch if possible
            self.get_node_from_snapshot(subj, pred)
            is_unique = dbutils.is_unique_predicate_value(pred)
            value = {pred: self.add_fact_id(
                self.typecast_input(subj, pred, objc),
                is_unique, pred == 'to:type')}
            if collection.find_path(self._stores, subj):
                collection.assign_value(self._stores, subj, value)
            elif collection.find_path(self._stores, subj_type[0], _id=subj):
                collection.assign_value(
                    self._stores, subj_type[0], value, _id=subj)
            else:
                self._stores['inmemory'][subj] = {'id': subj}
                collection.assign_value(self._stores, subj, value)

        # if subj_type_subclass isA to:Entity, subject can be found under _stores['snapshot']
        # assign the pred value under it
        # important thing here is the use of if, elif
        # the order ensures that we don't get any complex kind of assignment at this point
        # this is really simple property assertion, and that is why we don't get any
        # "if 'to:Property' in" kind of condition check anywhere
        elif 'to:Entity' in subj_type_subclass:
            # this means we have got a node.
            # get the subject from elasticsearch if possible
            self.get_node_from_snapshot(subj, pred)
            is_unique = dbutils.is_unique_predicate_value(pred)
            value = {pred: self.add_fact_id(
                self.typecast_input(subj, pred, objc),
                is_unique, pred == 'to:type')}
            if subj not in self._stores['snapshot']:
                self._stores['snapshot'][subj] = {
                    'id': subj
                }
            collection.assign_value(self._stores, subj, value)

    def get_snapshot(self):
        return self._stores['snapshot']

    def add_templatized_id_to_insert(self):
        prev_validation = self._validation
        self.turn_off_validation()
        for temp_id in self._new_templatized_id:
            if templatizedid.is_keep_id(temp_id):
                tofler_id = self._templatized_id_map[temp_id]
                self.add_input(tofler_id, 'to:templatizedId', temp_id)
        self._validation = prev_validation

    def commit(self):
        # create entries for any new storable templatized ids
        # we needed to call it seperately than handle_templatized_id,
        # as we want to call add_input(tofler_id, 'to:templatizedId', temp_id
        # once we have completed replacing temp_id with tofler_id
        try:
            self.add_templatized_id_to_insert()
            # self.add_superclasses_for_type_field(self._stores['snapshot'])
            # commit both the query list
            snapshot_dbutils.insert_into_snapshot(self._stores['snapshot'])
            eternity_dbutils.insert_into_eternity(self._input_list)
        finally:
            self.clean_up()

    def clean_up(self):
        self._input_list = []
        self._stores['snapshot'] = {}
        self._normalized_input = {}
        self._new_templatized_id = []
        locks.release_lock(self._locked_nodes)
        self._locked_nodes = []



'''ElasticSearch easy-access module

This module makes access to an ElasticSearch database very simple.
'''

from elasticsearch import Elasticsearch, helpers


class Elastic(object):

    def __init__(self, host, port):
        self._config = {
            'host': host,
            'port': port,
        }
        self._elasticdb = Elasticsearch(
            [{
                'host': self._config['host'],
                'port': self._config['port']
            }]
        )

    def get_connection(self):
        return self._elasticdb

    def put_bulk_data(self, index, mapping_name=None, data=[]):
        es = self.get_connection()
        actions = []
        response = None
        get_mapping_from_data = False
        if mapping_name is None:
            get_mapping_from_data = True
        for item in data:
            if get_mapping_from_data:
                action = {
                    '_index': index,
                    '_type': item[1],
                    '_source': item[0]
                }
                if 'id' in item[0]:
                    action['_id'] = item[0]['id']
            else:
                action = {
                    '_index': index,
                    '_type': mapping_name,
                    '_source': item
                }
                if 'id' in item:
                    action['_id'] = item['id']
            actions.append(action)
        if len(actions):
            response = helpers.bulk(es, actions)
        return response



from toflerdb.utils.common import Common


Q_BULK_UPLOAD_FACTS = 'BULK_UPLOAD_FACTS'
SNAPSHOT_INDEX = Common.get_config('elastic')['index']
SNAPSHOT_DOC_TYPE = Common.get_config('elastic')['doc_type']


class FACT_STATUS:
    ACTIVE = 'A'
    PENDING = 'P'
    DELETED = 'D'


class UPLOAD_TASK_STATUS:
    UPLOADED = 'U'
    PROGRESS = 'P'
    COMPLETED = 'C'
    FAILED = 'F'


UPLOAD_TASK_STATUS_LABEL = {
    UPLOAD_TASK_STATUS.UPLOADED: 'PENDING',
    UPLOAD_TASK_STATUS.PROGRESS: 'PROGRESS',
    UPLOAD_TASK_STATUS.COMPLETED: 'COMPLETED',
    UPLOAD_TASK_STATUS.FAILED: 'FAILED'
}



import redis


class Redis(object):

    def __init__(self, config):
        self._pool = redis.ConnectionPool(
            host=config['host'],
            port=config['port'],
            db=0
            )

    def set(self, key, value):
        r = redis.Redis(connection_pool=self._pool)
        return r.set(key, value)

    def get(self, key):
        r = redis.Redis(connection_pool=self._pool)
        return r.get(key)

    def incr(self, key):
        r = redis.Redis(connection_pool=self._pool)
        return r.incr(key)

    def get_connection(self):
        return redis.Redis(connection_pool=self._pool)



import os
from pkg_resources import resource_filename
import argparse
from toflerdb.core import api


class Sampler(object):

    DATA_PATH = 'resources/data/'

    def __init__(self):
        self.DATA_PATH = resource_filename('toflerdb', self.DATA_PATH)
        self.ONTOLOGY_FILE = os.path.join(
            self.DATA_PATH, 'ussec.ontology.rdf')
        self.FACTS_FILE = os.path.join(
            self.DATA_PATH, 'ussec.sample-facts.rdf')

    def create_ontology(self):
        print "Creating Ontology ..."
        api.insert_ontology(file=self.ONTOLOGY_FILE, validation=False)
        print "Done ."

    def create_facts(self):
        print "Inserting Facts ..."
        api.insert_facts(
            file=self.FACTS_FILE, author='toflerdb', show_progress=True)
        print "Done ."

    def run(self):
        self.create_ontology()
        self.create_facts()


def parse_arguments():
    parser = argparse.ArgumentParser()
    parser.add_argument('--config', help='Config file')
    return parser.parse_args()


def run_sample():
    args = parse_arguments()
    prev_conf = ''
    if args.config:
        prev_conf = os.getenv('TOFLERDB_CONF', '')
        os.environ['TOFLERDB_CONF'] = args.config

    sampler = Sampler()
    sampler.run()
    os.environ['TOFLERDB_CONF'] = prev_conf


if __name__ == '__main__':
    run_sample()



import json
# from toflerdb.utils.common import Common
from toflerdb.dbutils import snapshot as snapshot_dbutils
from toflerdb.utils import exceptions
from basehandler import BaseHandler


class GeneralSearchHandler(BaseHandler):

    def post(self):
        self.get()

    def get(self):
        request_body = self.request.body
        if request_body is None:
            return []
        response = []
        try:
            request_body = json.loads(request_body)
        except:
            print "Error processing request"
            response = []
        search_text = request_body.get('q', None)
        try:
            response = snapshot_dbutils.general_search(search_text)
            self.apiwrite(response)
        except exceptions.ToflerDBException, e:
            self.apiwrite(str(e), status=False)
        except Exception, e:
            # Common.get_logger().error(str(e))
            self.apiwrite('Something went wrong', status=False)



'''
Steps to be followed
to setup the toflerdb project from scratch
OR
to clean up the toflerdb project to its empty state
    Step1: Delete toflerdb_snapshot index if exists
    Step2: Create intial mapping/analyzer for type and label
    Step3: Empty toflerdb_eternity & toflerdb_ontology table
    Step4: Import the common ontology, like Property, Entity
'''

import os
from toflerdb.utils.common import Common
from toflerdb import config
from toflerdb.core import api
from pkg_resources import resource_filename


class Bootstrapper(object):

    SCHEMA_PATH = 'resources/dbpatches/'
    ONTOLOGY_PATH = 'resources/data/to.common.ontology.rdf'

    def __init__(self):
        self.ONTOLOGY_PATH = resource_filename('toflerdb', self.ONTOLOGY_PATH)
        self.SCHEMA_PATH = resource_filename('toflerdb', self.SCHEMA_PATH)
        self.WORKER_FILE = os.path.join(
            self.SCHEMA_PATH, 'toflerdb_worker.sql')
        self.ONTOLOGY_FILE = os.path.join(
            self.SCHEMA_PATH, 'toflerdb_ontology.sql')
        self.ETERNITY_FILE = os.path.join(
            self.SCHEMA_PATH, 'toflerdb_eternity.sql')
        self.INDEX_MAPPING_FILE = os.path.join(
            self.SCHEMA_PATH, 'toflerdb_snapshot.es')
        self.DOCTYPE_MAPPING_FILE = os.path.join(
            self.SCHEMA_PATH, 'toflerdb_snapshot.node.es')

    def log(self, msg):
        print msg

    def execute_multi_query(self, query):
        queries = query.split(';')
        for q in queries:
            q = q.strip()
            if q:
                Common.execute_query(q)

    def create_sql_tables(self):
        query = open(self.WORKER_FILE, 'r').read()
        self.execute_multi_query(query)
        self.log("Creating Ontology Table ...")
        query = open(self.ONTOLOGY_FILE, 'r').read()
        self.execute_multi_query(query)
        self.log("Creating Eternity Table ...")
        query = open(self.ETERNITY_FILE, 'r').read()
        self.execute_multi_query(query)

    def create_es_mappings(self):
        self.log("Creating Index Mapping ...")
        es = Common.get_elasticsearch_connection()
        es.indices.delete(
            index=config.SNAPSHOT_INDEX, ignore=[400, 404])
        mapping = open(self.INDEX_MAPPING_FILE).read()
        es.indices.create(
            index=config.SNAPSHOT_INDEX, ignore=400, body=mapping)
        mapping = open(self.DOCTYPE_MAPPING_FILE).read()
        es.indices.put_mapping(
            index=config.SNAPSHOT_INDEX,
            doc_type=config.SNAPSHOT_DOC_TYPE,
            ignore=400, body=mapping)

    def reboot(self):
        self.create_es_mappings()
        self.create_sql_tables()
        self.log("Inserting Base Ontology ...")
        prefilled_ontology = [
            ('to:type', 'to:subClassOf', 'to:Property'),
            ('to:type', 'to:domain', 'to:Entity'),
            ('to:type', 'to:domain', 'to:ComplexProperty'),
            ('to:type', 'to:domain', 'to:ComplexRelationalProperty'),
            ('to:type', 'to:range', 'to:Token'),
            ('to:templatizedId', 'to:subClassOf', 'to:Property'),
            ('to:templatizedId', 'to:domain', 'to:Entity'),
            ('to:templatizedId', 'to:domain', 'to:ComplexProperty'),
            ('to:templatizedId', 'to:domain', 'to:ComplexRelationalProperty'),
            ('to:templatizedId', 'to:range', 'to:ID'),
            ('to:templatizedId', 'to:isUnique', 'True'),
            ('to:label', 'to:subClassOf', 'to:Property'),
            ('to:label', 'to:domain', 'to:Entity'),
            ('to:label', 'to:range', 'to:String')
        ]
        query = """
            INSERT INTO toflerdb_ontology(subject, predicate, value)
            VALUES(%s, %s, %s)
        """
        for row in prefilled_ontology:
            Common.execute_query(query, row)

        api.insert_ontology(
            file=self.ONTOLOGY_PATH,
            validation=False)


def run_bootstrap():
    bootstrapper = Bootstrapper()
    bootstrapper.reboot()


if __name__ == '__main__':
    run_bootstrap()



from toflerdb.utils.common import Common
from toflerdb.utils import templatizedid
from toflerdb.config import FACT_STATUS


def convert_elements_to_string(subj, pred, objc):
    if subj is not None:
        subj = str(subj)
    if pred is not None:
        pred = str(pred)
    if objc is not None:
        objc = str(objc)

    return (subj, pred, objc)


def tuple_exists_in_eternity(subj, pred, objc):
    (subj, pred, objc) = convert_elements_to_string(subj, pred, objc)
    # print type(subj), type(pred), type(objc)
    # print subj, pred, objc
    query = """
        SELECT subject FROM toflerdb_eternity WHERE subject = %s AND predicate
        = %s AND (object = %s OR value = %s) AND status != %s
    """
    response = Common.execute_query(query, (
        subj, pred, objc, objc, FACT_STATUS.DELETED))
    if len(response):
        return True
    return False


def insert_into_eternity(inputs):
    if not inputs:
        return
    query = """
        INSERT INTO toflerdb_eternity(fact_id, subject, predicate, object,
         value, uid, prev) VALUES
    """
    query_data = []
    query_clauses = []
    for item in inputs:
        query_clauses.append('%s, %s, %s, %s, %s, %s, %s')
        query_data += [item['fact_id'], item['subject'],
                       item['predicate'], item['object'], item['value'],
                       item['author'], item['prev']]
    query_clauses_str = '), ('.join(query_clauses)
    query += '(' + query_clauses_str + ')'
    Common.execute_query(query, tuple(query_data))


def get_fact_ids(fact_tuples, author=None):
    query = """
        SELECT fact_id FROM toflerdb_eternity WHERE
    """
    placeholder = []
    query_data = []
    for row in fact_tuples:
        row = list(row)
        subj = row[0]
        if templatizedid.is_templatized_id(subj):
            subj = templatizedid.append_userid(subj, author)
            subj = get_id_by_templatized_id(subj)
        if not subj:
            continue

        placeholder.append('(subject=%s AND predicate=%s AND \
                             (object=%s OR value=%s))')


        (subj, row[1], row[2]) = convert_elements_to_string(subj, row[1], row[2])
        query_data += [subj, row[1], row[2], row[2]]

    if not len(placeholder):
        return []

    placeholder_str = ' OR '.join(placeholder)
    query += placeholder_str
    response = Common.execute_query(query, tuple(query_data))
    retval = []
    for res in response:
        retval.append(res['fact_id'])

    return retval


def get_fact_status(fact_ids):
    query = """
        SELECT fact_id, status FROM toflerdb_eternity WHERE fact_id IN
    """
    placeholder_str = ', '.join(['%s' for _ in fact_ids])
    query = '%s (%s)' % (query, placeholder_str)
    query_data = tuple(fact_ids)
    response = Common.execute_query(query, query_data)
    retval = {res['fact_id']: res['status'] for res in response}

    return retval


def is_all_eternity_facts_deleted(fact_ids):
    statuses = get_fact_status(fact_ids)
    for fid in fact_ids:
        if statuses.get(fid) != FACT_STATUS.DELETED:
            return False

    return True


def is_none_eternity_facts_deleted(fact_ids):
    statuses = get_fact_status(fact_ids)
    for fid in fact_ids:
        if statuses.get(fid) == FACT_STATUS.DELETED:
            return False

    return True


def get_fact_tuples_by_fact_ids(fact_ids):
    if not isinstance(fact_ids, list):
        fact_ids = [fact_ids]

    placeholder_str = ', '.join(['%s' for _ in fact_ids])
    query = """
        SELECT subject, predicate, object, value FROM toflerdb_eternity
        WHERE fact_id IN (%s)
    """ % placeholder_str
    query_data = tuple(fact_ids)
    response = Common.execute_query(query, query_data)
    return [(res['subject'], res['predicate'], res['object'] or res['value'])
            for res in response]


def get_id_by_templatized_id(temp_id):
    query = """
        SELECT subject FROM toflerdb_eternity WHERE predicate = %s AND
        value = %s
    """
    query_data = ('to:templatizedId', temp_id)
    response = Common.execute_query(query, query_data)
    retval = None
    if len(response):
        retval = response[0]['subject']

    return retval


def is_any_type_facts(fact_ids):
    if not isinstance(fact_ids, list):
        fact_ids = [fact_ids]

    placeholder_str = ', '.join(['%s' for _ in fact_ids])
    query = """
        SELECT fact_id FROM toflerdb_eternity WHERE fact_id IN (%s)
        AND predicate = 'to:type'
    """ % placeholder_str
    query_data = tuple(fact_ids)
    response = Common.execute_query(query, query_data)
    return len(response) > 0



# *-* coding: utf-8 -*-
from setuptools import setup, find_packages
from os.path import join, dirname

VERSION_FILE = open(join(dirname(__file__), 'VERSION'))
__versionstr__ = VERSION_FILE.read().strip()
VERSION_FILE.close()

REQUIREMENTS = open(join(dirname(__file__), 'requirements.txt'))
INSTALL_REQUIRES = REQUIREMENTS.read().split('\n')
REQUIREMENTS.close()

f = open(join(dirname(__file__), 'DESCRIPTION'))
long_description = f.read().strip()
f.close()

CONSOLE_SCRIPTS = [
    'toflerdb-server=toflerdb.apiserver:run_apiserver',
    'toflerdb-dbinit=toflerdb.scripts.bootstrap:run_bootstrap',
    'toflerdb-run-sample=toflerdb.scripts.sample:run_sample',
    'toflerdb-generate-config=toflerdb.scripts.sampleconfig:run',
]


setup(
    name='toflerdb',
    version=__versionstr__,
    description='ToflerDB',
    long_description=long_description,
    url='http://www.toflerdb.com',
    author='Team Tofler',
    author_email='devteam@tofler.in',
    license='GNU AGPLv3',
    package_dir={'': './'},
    packages=find_packages('./'),
    install_requires=INSTALL_REQUIRES,
    dependency_links=[
        'http://cdn.mysql.com/Downloads/Connector-Python/mysql-connector-python-1.2.3.zip#md5=6d42998cfec6e85b902d4ffa5a35ce86'
    ],
    include_package_data=True,
    package_data={'toflerdb': [
        'resources/data/*', 'resources/dbpatches/*']
    },
    classifiers=[
        'Development Status :: 4 - Beta',
        'Intended Audience :: Developers',
        'Programming Language :: Python :: 2.7',
        'Operating System :: OS Independent'
    ],
    entry_points={
        'console_scripts': CONSOLE_SCRIPTS,
    }
    )



import logging
from .config import Configuration
from .db import Database
from .elastic import Elastic
from .logger import Logger, ElasticLogHandler, FORMAT
from .idmanagement import IDManager
from .cache import Redis


class Common(object):

    configuration = None
    database = None
    elastic = None
    logger = None
    idmanager = None
    redis = None

    @classmethod
    def load_config(cls, location=None):
        cls.configuration = Configuration(location)

    @classmethod
    def get_config(cls, section):
        if cls.configuration is None:
            cls.load_config()
        return cls.configuration.get_config(section)

    @classmethod
    def _get_database(cls):
        if cls.configuration is None:
            cls.load_config()
        if cls.database is None:
            cls.database = Database(
                pool_name='toflerdb',
                config=cls.configuration.get_config('db')
            )
        return cls.database

    @classmethod
    def execute_query(cls, query, params=None, **kwargs):
        if cls.database is None:
            cls._get_database()
        return cls.database.execute_query(query, params, **kwargs)

    @classmethod
    def _create_elastic_object(cls):
        if cls.elastic is None:
            if cls.configuration is None:
                cls.load_config()
            # Check if we have a configuration for es
            config = cls.configuration.get_config('elastic')
            cls.elastic = Elastic(config['host'], config['port'])

    @classmethod
    def get_elasticsearch_connection(cls):
        if cls.elastic is None:
            cls._create_elastic_object()
        return cls.elastic.get_connection()

    @classmethod
    def get_elasticsearch(cls):
        if cls.elastic is None:
            cls._create_elastic_object()
        return cls.elastic

    @classmethod
    def get_logger(cls, log_level=logging.INFO):
        if cls.logger is None:
            formatter = logging.Formatter(FORMAT)
            cls._create_elastic_object()
            elh = ElasticLogHandler(cls.elastic)
            elh.setFormatter(formatter)
            logging.setLoggerClass(Logger)
            logger = logging.getLogger('tofler')
            logger.setLevel(log_level)
            logger.addHandler(elh)
            cls.logger = logger
            return logger
        else:
            return cls.logger

    @classmethod
    def _get_idmanager(cls):
        if cls.idmanager is None:
            cls._get_database()
            cls.idmanager = IDManager(cls.database, cls.get_logger())
        return cls.idmanager

    @classmethod
    def generate_id(cls, namespace=None):
        idmanager = cls._get_idmanager()
        return idmanager.generate_new_id(namespace)

    @classmethod
    def _make_redis(cls):
        if cls.redis is not None:
            return
        if cls.configuration is None:
            cls.load_config()
        redis_config = cls.get_config('redis')
        cls.redis = Redis(config=redis_config)

    @classmethod
    def get_cache_connection(cls):
        if cls.redis is None:
            cls._make_redis()
        return cls.redis.get_connection()



import argparse
import tornado.ioloop
import tornado.web
from handlers.queryhandler import QueryHandler
from handlers.generalsearchhandler import GeneralSearchHandler
from handlers.uploadontologyhandler import UploadOntologyHandler
from handlers.uploadfactshandler import UploadFactsHandler


def parse_arguments():
    parser = argparse.ArgumentParser()
    parser.add_argument(
        '--port',
        help='Port on which dbserver will run',
        type=int, default=8888
    )
    return parser.parse_args()


def make_app():

    query_endpoints = [
        (r'/query', QueryHandler),
        (r'/search', GeneralSearchHandler),
    ]
    user_endpoints = [
    ]
    upload_endpoints = [
        (r'/upload/ontology', UploadOntologyHandler),
        (r'/upload/facts', UploadFactsHandler),
    ]

    handlers = []
    handlers += query_endpoints
    handlers += user_endpoints
    handlers += upload_endpoints

    return tornado.web.Application(handlers)


def main():
    args = parse_arguments()
    print "Starting dbserver on port %s" % args.port
    app = make_app()
    app.listen(args.port)
    tornado.ioloop.IOLoop.current().start()


if __name__ == "__main__":
    main()



from toflerdb.dbutils import dbutils


def get_all_namespaces():
    return dbutils.get_all_namespaces()


def get_all_entities_of_namespace(namespace):
    return dbutils.get_all_entities_of_namespace(namespace)


def get_all_propeties(elem):
    return dbutils.get_all_propeties(elem)



'''ToflerDB Configuration Management Module

This purpose of this module is to provide easy access to Tofler's
configuration file located either at the default location or at
a location as specified in a call.

This also provides a mechanism to alter an existing configuration

By default, the Configuration class will look for configuration
files in the following order:
    1. the location file (if) specified in the constructor
    2. file specified in the TOFLERDB_CONF environment variable
    3. $HOME/.tofler.conf

The configuration file must be a json encoded dictionary
with each "section" being a top-level key in the dictionary
'''

import os
import json


def get_base_config():
    return {}


class ConfigFileNotFoundError(ValueError):
    '''Raised when the specified configuration file is not found'''


class ConfigurationParseError(ValueError):
    '''Raised when the configuration could not be parsed correctly'''


class ConfigurationAlreadyLoadedError(ValueError):
    '''Raised when there is an attempt to re-load a configuration'''


class SectionNotFoundError(ValueError):
    '''Raised when a requested section is not found in the configuration'''


class Configuration(object):
    '''This class contains a single instance of a configuration'''

    def __init__(self, location=None):
        self._config = {}
        self._config_file = None
        if location is None:
            location = os.getenv('TOFLERDB_CONF')
            if location is None or location.strip() == '':
                home = os.getenv('HOME')
                location = os.path.join(home, '.toflerdb.conf')
        if not os.path.exists(location):
            # raise ConfigFileNotFoundError('%s was not found' % location)
            pass
        else:
            self._config_file = location
        self._load_config()

    def _load_config(self):
        '''Private method to load configuration file'''
        self._config = get_base_config()

        if self._config_file:
            config_file = open(self._config_file, 'r')
            text = config_file.read()
            config_file.close()
            try:
                self._config.update(json.loads(text))
            except ValueError:
                raise ConfigurationParseError('Error parsing configuration')

    def get_config(self, section):
        '''Public method to access configuration for a section'''

        if section not in self._config:
            raise SectionNotFoundError(
                '%s not found in configuration' % section)
        return self._config[section]
